from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from manim import *


class TestTextManimAnimation(BaseManimAnimation):
    """
    Animation just for testing with texts.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, text: str, duration: float, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid. The 
        'text' parameter is limited to 30 characters.
        """
        # Check and validate all parameters
        parameters = {}

        # Empty, for testing

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename
    
    def animate(self):
        """
        This code will generate the manim animation and belongs to the
        Scene manim object.
        """
        text = Text('Texto de prueba')

        self.play(
            AnimationGroup(
                text.animate(run_time = 1).shift(RIGHT),
                text.animate.rotate(PI / 8),
                text.animate(run_time = 0.8).scale(2)
            ),
            run_time = 2
        )

        self.wait(1)


from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_general_utils.checker.url import is_google_drive_url
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.file.checker import file_exists
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.path import create_file_abspath


class Resource:
    """
    Class to simplify the way we work with a local resource or with one that
    is stored in Google Drive.
    """
    @classmethod
    def get(cls, filename_or_google_drive_url: str, output_filename: str = None):
        """
        This method has been built to be used internally to obtain
        the resources this software needs always in the same way
        and to be able to speed up the process by avoiding external
        downloads as much as possible.
        
        If the provided 'filename_or_google_drive_url' is a Google
        Drive url, the first time a resource is downloaded from the
        given 'google_drive_url', if 'output_filename' provided, will 
        be stored locally with that filename. Next time, as it will
        be found locally, it will be returned and we will have not to
        wait for the download.
        """
        if not filename_or_google_drive_url:
            raise Exception('No "filename_or_google_drive_url" parameter provided.')
        
        if output_filename and file_exists(output_filename):
            return output_filename
        
        if is_google_drive_url(filename_or_google_drive_url):
            # TODO: Check that abspath / output_filename is valid

            # We force to create all folders if they don't exist
            create_file_abspath(f'{get_project_abspath()}{output_filename}')

            output_filename = GoogleDriveResource(filename_or_google_drive_url).download(output_filename)
        else:
            if not file_exists(filename_or_google_drive_url):
                raise Exception(f'The provided "filename" parameter {filename_or_google_drive_url} does not exist.')
            
            # TODO: I can copy it to be found as output_filename,
            # but I don't want by now.
            # if filename_or_google_drive_url != output_filename:
            #     copy_file(filename_or_google_drive_url, output_filename)
            output_filename = filename_or_google_drive_url

        return output_filename


# TODO: Remove this below when refactored
def get_resource(google_drive_url: str, output_filename: str = None):
    """
    This method has been built to be used internally to obtain
    the resources this software needs. This method will look
    for the 'output_filename' if provided, and if it exist
    locally stored, it will return that filename. If it doesn't
    exist or the 'output_filename' is not provided, it will 
    download the resource from the given 'google_drive_url',
    store it locally and return it.

    The first time a resource is downloaded from the given
    'google_drive_url', if 'output_filename' provided, will
    be stored locally with that filename. Next time, as it 
    will be found locally, it will be returned and we will
    have not to wait for the download.

    This method has been created to optimize the resource
    getting process.

    @param
        **google_drive_url**
        The Google Drive url in which we can found (and
        download) the resource.

    @param
        **output_filename**
        The filename of the resource we want to get, that
        must be a relative path from the current project
        abspath.
    """
    if not google_drive_url:
        raise Exception('No "google_drive_url" parameter provided.')
    
    output_abspath = create_temp_filename('tmp_resource.mp3')
    
    if output_filename:
        # We try to find it and return or we create the folder if doesn't exist
        output_abspath = get_project_abspath() + output_filename

        if file_exists(output_abspath):
            return output_abspath
        
        # We force to create all folders
        create_file_abspath(output_abspath)
    
    # We download the file and store it locally
    return GoogleDriveResource(google_drive_url).download(output_abspath)


    # Has 'output_filename'
        # Resource is Google Drive url
            # => We check if 'output_filename' and return if yes
            # => We check if 'output_filename' and download if not as it
        # Resource is filename (?)
            # => We check if 'output_filename' exist and return it (!)
            # => We copy the filename as 'output_filename' and return it
    # Doesn't have 'output_filename'
        # Resource is Google Drive url
            # => We download it and return the temporary file downloaded
        # Resource is filename
            # => We check if existing and return the filename as it is
    


from yta_general_utils.file.writer import write_file
from yta_general_utils.file.reader import read_json_from_file
from yta_general_utils.programming.path import get_project_abspath

import json


CONFIG_MANIM_ABSPATH = get_project_abspath() + 'manim_parameters.json'

class ManimConfig:
    @classmethod
    def write(cls, json_data):
        """
        Writes the the provided 'json_data' manim configuration
        in the configuration file so the parameters could be read
        later by the manim engine. This is the way to share 
        parameters to the process.
        """
        # TODO: Check that 'json_data' is valid and well-formatted
        write_file(json.dumps(json_data, indent = 4), CONFIG_MANIM_ABSPATH)

    @classmethod
    def read(cls):
        """
        Read the configuration file and return it as a json
        object.
        """
        return read_json_from_file(CONFIG_MANIM_ABSPATH)



# TODO: Remove this below when unneded
def write_manim_config_file(json_data):
    """
    Writes in the configuration file that we use to share
    parameters with manim software. This is the way to 
    share parameters to the process.
    
    TODO: I would like to be able to handle manim through 
    python code directly and not an external process I run,
    but for now this is working.
    """
    # We serialize json to str
    json_object_str = json.dumps(json_data, indent = 4)
    
    write_file(json_object_str, CONFIG_MANIM_ABSPATH)

def read_manim_config_file():
    """
    Reads the configuration file and returns it as a json
    object.
    """
    return read_json_from_file(CONFIG_MANIM_ABSPATH)


from yta_multimedia.audio.parser import AudioParser
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.audio.converter import AudioConverter
from yta_general_utils.file.remover import delete_file
from df.enhance import enhance, init_df, load_audio, save_audio


class AudioNoise:
    """
    Class to simplify and encapsulate the code related with audio noise.
    """
    def remove(audio, output_filename: str = None):
        """
        Remove the noise from the provided audio and, if 'output_filename'
        is provided, the audio without noise is written localy with that
        filename.
        """
        # TODO: This fails when .mp3 is used, so we need to transform into wav.
        # TODO: Output file must be also wav
        # TODO: What about audioclip instead of audiofile? Is it possible? (?)
        # Based on this (https://medium.com/@devesh_kumar/how-to-remove-noise-from-audio-in-less-than-10-seconds-8a1b31a5143a)
        # https://github.com/Rikorose/DeepFilterNet
        # TODO: This is failing now saying 'File contains data in an unknon format'...
        # I don't know if maybe some library, sh*t...
        # Load default model
        
        # If it is not an audio filename I need to create it to be able to
        # work with (TODO: Check how to turn into same format as when readed)
        # TODO: Refactor these below to accept any audio, not only filename
        tmp_audio_filename = create_temp_filename('audio_temp.wav')

        audio = AudioParser.to_audiosegment(audio)
        _, tmp_audio_filename = AudioConverter.to_wav(audio, tmp_audio_filename) 

        # TODO: This was done before because the parameter was only
        # a filename and now I'm accepting other audio types
        # if audio_filename.endswith('.mp3'):
        #     # TODO: Maybe it is .wav but not that format...
        #     mp3_to_wav(audio_filename, TMP_WAV_FILENAME)
        #     audio_filename = TMP_WAV_FILENAME

        model, df_state, _ = init_df()
        audio, _ = load_audio(tmp_audio_filename, sr = df_state.sr())
        # Remove the noise
        enhanced = enhance(model, df_state, audio)

        if not output_filename:
            output_filename = create_temp_filename('noise_removed.wav')
        # TODO: Check temp filename extension is valid

        save_audio(output_filename, enhanced, df_state.sr())

        try:
            delete_file(tmp_audio_filename)
        except:
            pass

    # TODO: Create a 'generate' noise method


from manim.mobject.opengl.opengl_image_mobject import OpenGLImageMobject as OriginalOpenGLImageMobject
from pathlib import Path
from PIL import Image
from typing import Union

import numpy as np


class OpenGLImageMobject:
    """
    Custom class created to use its 'init' method to solve a problem with
    filename passed as a parameter, that is not correctly parsed by the
    original manim OpenGLImageMobject class.

    This class is not instantiable, just use OpenGLImageMobject.init(params)
    to obtain a real manim OpenGLImageMobject instance.
    """
    # TODO: I would like to use it as OpenGLImageMobject and that init method
    # return an instance of the original manim OpenGLImageMobject and not this
    # one, but as I don't know exactly how to do it I'm creating this static
    # 'init' method to achieve it. Improve it if possible, thanks.
    @staticmethod
    def init(
        filename_or_array: Union[str, Path, np.ndarray],
        width: float = None,
        height: float = None,
        image_mode: str = "RGBA",
        resampling_algorithm: int = Image.Resampling.BICUBIC,
        opacity: float = 1,
        gloss: float = 0,
        shadow: float = 0,
        **kwargs
    ):
        """
        This method returns an instance of a manim OpenGLImageMobject.
        """
        if isinstance(filename_or_array, str):
            filename_or_array = np.asarray(Image.open(filename_or_array))

        return OriginalOpenGLImageMobject(filename_or_array, width, height, image_mode, resampling_algorithm, opacity, gloss, shadow, **kwargs)

    def __init__(self):
        """
        DO NOT USE THIS METHOD. This class is not instantiable.
        """
        raise Exception('Sorry, this class is not instantiable, just use the "init" method to obtain a OpenGLImageMobject instance.')


from yta_multimedia.audio.parser import AudioParser
from yta_general_utils.checker.type import variable_is_positive_number
from pydub import silence


class AudioSilence:
    """
    Class to simplify and encapsulate the interaction with audio silences.
    """
    @classmethod
    def detect(cls, audio, minimum_silence_ms: int):
        """
        Detect the silences of a minimum of 'minimum_silence_ms' milliseconds
        time and returns an array containing tuples with the start and the 
        end of the silence moments.
        """
        audio = AudioParser.to_audiosegment(audio)

        if not minimum_silence_ms:
            minimum_silence_ms = 250

        if not variable_is_positive_number(minimum_silence_ms):
            raise Exception('The provided "minimum_silence_ms" is not a positive number.')

        dBFS = audio.dBFS
        # TODO: Why '- 16' (?) I don't know
        silences = silence.detect_silence(audio, min_silence_len = minimum_silence_ms, silence_thresh = dBFS - 16)

        # [(1.531, 1.946), (..., ...), ...] in seconds
        return [((start / 1000), (stop / 1000)) for start, stop in silences]


from yta_general_utils.image.parser import ImageParser
from yta_general_utils.checker.type import variable_is_positive_number
from moviepy.editor import ImageClip
from yta_multimedia.video.parser import VideoParser
from typing import Union
import numpy as np


def generate_video_from_image(image, duration: float = 1, output_filename: Union[str, None] = None):
    """
    Receives an image and creates an ImageClip of 'duration' seconds.
    It will be also stored as a file if 'output_filename' is provided.
    """
    if not variable_is_positive_number(duration):
        raise Exception(f'The provided "duration" parameter {str(duration)} is not a positive number.')

    if not isinstance(image, ImageClip):
        video = ImageClip(ImageParser.to_numpy(image)).set_fps(60).set_duration(duration)

    if output_filename:
        video.write_videofile(output_filename)

    return video

def is_video_transparent(video):
    """
    Checks if the first frame of the mask of the given 'video'
    has, at least, one transparent pixel.
    """
    # We need to detect the transparency from the mask
    video = VideoParser.to_moviepy(video, has_mask = True)

    # We need to find, by now, at least one transparent pixel
    # TODO: I would need to check all frames to be sure of this above
    return np.any(video.mask.get_frame(t = 0) == 1)


def zoom_in_t_func(zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video be zoomed in the 'zoom_ratio' zoom all the time.
    """
    return 1 + zoom_ratio

def zoom_out_t_func(zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video be zoomed out the 'zoom_ratio' zoom all the time.
    """
    return 1 - zoom_ratio

def linear_zoom_in_t_func(t, duration, zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video zoom in linearly to the 'zoom_ratio' zoom.
    """
    # TODO: Check 'zoom_ratio' is valid

    return 1 + zoom_ratio * (t / duration)

def linear_zoom_out_t_func(t, duration, zoom_ratio: float = 0.2):
    """
    Function that returns the resize value to make a moviepy
    video zoom out linearly to the 'zoom_ratio' zoom.
    """
    # TODO: Check 'zoom_ratio' is valid

    return 1 + zoom_ratio * (duration - (t / duration))

def linear_zoom_transition_t_func(t, duration, zoom_start: float = 1, zoom_end: float = 0.8):
    """
    Function that returns the resize value to make a moviepy
    video zoom in or out from the 'zoom_start' to the 'zoom_end'
    in the provided 'duration' time. If 'zoom_start' is greater
    than 'zoom_end' that will be a zoom out.
    """
    return zoom_start + (zoom_end - zoom_start) * (t / duration)

def logaritmic_zoom_transition_t_func(t, duration, zoom_start: float = 1, zoom_end: float = 0.8):
    # TODO: This has not been tested...
    import math

    zoom_start = math.log(zoom_start)
    zoom_end = math.log(zoom_end)

    # Interpolación logarítmica
    return math.exp(zoom_start + (zoom_end - zoom_start) * (t / duration))




from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.custom.utils import get_greenscreen_details
from yta_multimedia.alphascreen.masked_clip_creator import MaskedClipCreator
from yta_multimedia.greenscreen.enums import GreenscreenType
from yta_general_utils.checker.type import variable_is_type
from moviepy.editor import vfx, VideoClip, VideoFileClip
from typing import Union


class VideoGreenscreen:
    """
    Class representing a Video with some greenscreen regions on it
    that can be used to place other resources (images or videos)
    fitting those regions while this greenscreen video is displayed.

    This class is working as if the greenscreen regions in the image
    were static, so if it moves this won't work properly as it is 
    not mapping the region for all the frames, just for the first 
    one.

    TODO: Improve this by autodetecting all the greenscreen regions
    for each frame and storing them somewhere.
    """
    greenscreen: GreenscreenDetails = None
    """
    This parameter keeps the information about the greenscreen 
    regions that the video has, including their corner coordinates,
    the width, the height and the green color to apply the mask.
    """

    def __init__(self, greenscreen: Union[GreenscreenDetails, str]):
        # TODO: Enhance this by detecting greenscreens for each frame
        if variable_is_type(greenscreen, str):
            # We need to automatically detect greenscreen details
            greenscreen = get_greenscreen_details(greenscreen, GreenscreenType.VIDEO)

        self.greenscreen = greenscreen

        # TODO: Do this here to be able to use it in the masked_clip_creator
        TMP_FILENAME = self.greenscreen.get_filename()
        # I consider the same greenscreen rgb color for all areas
        greenscreen_clip = VideoFileClip(TMP_FILENAME).fx(vfx.mask_color, color = self.greenscreen.greenscreen_areas[0].rgb_color, thr = 100, s = 5)

        regions = [gsa.region for gsa in self.greenscreen.greenscreen_areas]

        self.masked_clip_creator = MaskedClipCreator(regions, greenscreen_clip)
       
    def from_image_to_image(self, image, output_filename: str):
        """
        Receives an 'image', places it into the greenscreen and generates
        an image with the first clip that is stored locally as
        'output_filename' if provided.
        """
        # TODO: This is not returning RGBA only RGB
        return self.masked_clip_creator.from_image_to_image(image, output_filename)
    
    def from_images_to_image(self, images, output_filename: str):
        return self.masked_clip_creator.from_images_to_image(images, output_filename)
    
    def from_image_to_video(self, image, duration: float, output_filename: str):
        """
        Receives an 'image', places it into the greenscreen and generates
        a video of 'duration' seconds of duration that is returned. This method
        will store locally the video if 'output_filename' is provided.
        """
        return self.masked_clip_creator.from_image_to_video(image, duration, output_filename)
    
    def from_images_to_video(self, images, duration: float, output_filename: str = None):
        return self.masked_clip_creator.from_images_to_video(images, duration, output_filename)
    
    def from_video_to_video(self, video: Union[str, VideoClip], output_filename: str = None):
        """
        Inserts the provided 'video' in the greenscreen and returns the
        CompositeVideoClip that has been created. If 'output_filename' 
        provided, it will be written locally with that file name.

        The provided 'video' can be a filename or a moviepy video clip.
        """
        return self.masked_clip_creator.from_video_to_video(video, output_filename)
    
    def from_videos_to_video(self, videos: list[Union[str, VideoClip]], output_filename: str = None):
        """
        Puts the provided 'videos' inside the greenscreen region by
        applying a mask, cropping the videos if necessary and rescaling
        them, also positioning to fit the region and returns the 
        CompositeVideoClip created.

        Videos can be longer or shorter than greenscreen clip. By now
        we are making that all videos fit the greenscreen duration. 
        That is achieved by enlarging or shortening them if necessary.
        Thats why results could be not as expected.

        TODO: Please, build some different strategies to apply here.
        """
        return self.masked_clip_creator.from_videos_to_video(videos, output_filename)


from yta_multimedia.video.edition.effect.moviepy.position.utils import get_center_x, get_center_y
from random import randrange, choice as randomchoice
from enum import Enum


class ScreenPosition(Enum):
    OUT_TOP_LEFT = 'out_top_left'
    """
    Out of the screen, on the top left corner, just one pixel
    out of bounds.
    """
    IN_EDGE_TOP_LEFT = 'in_edge_top_left'
    """
    The center of the video is on the top left corner, so only
    the bottom right quarter part of the video is shown (inside
    the screen).
    """
    TOP_LEFT = 'top_left'
    """
    The video is completely visible, just at the top left 
    corner of the screen.
    """
    OUT_TOP = 'out_top'
    IN_EDGE_TOP = 'in_edge_top'
    TOP = 'top'
    OUT_TOP_RIGHT = 'out_top_right'
    IN_EDGE_TOP_RIGHT = 'in_edge_top_right'
    TOP_RIGHT = 'top_right'
    CENTER = 'center'
    OUT_RIGHT = 'out_right'
    IN_EDGE_RIGHT = 'in_edge_right'
    RIGHT = 'right'
    OUT_BOTTOM_RIGHT = 'out_bottom_right'
    IN_EDGE_BOTTOM_RIGHT = 'in_edge_bottom_right'
    BOTTOM_RIGHT = 'bottom_right'
    OUT_BOTTOM = 'out_bottom'
    IN_EDGE_BOTTOM = 'in_edge_bottom'
    BOTTOM = 'bottom'
    OUT_BOTTOM_LEFT = 'out_bottom_left'
    IN_EDGE_BOTTOM_LEFT = 'in_edge_bottom_left'
    BOTTOM_LEFT = 'bottom_left'
    OUT_LEFT = 'out_left'
    IN_EDGE_LEFT = 'in_edge_left'
    LEFT = 'left'

    HALF_TOP = 'half_top'
    HALF_TOP_RIGHT = 'half_top_right'
    HALF_RIGHT = 'half_right'
    HALF_BOTTOM_RIGHT = 'half_bottom_right'
    HALF_BOTTOM = 'half_bottom'
    HALF_BOTTOM_LEFT = 'half_bottom_left'
    HALF_LEFT = 'half_left'
    HALF_TOP_LEFT = 'half_top_left'

    QUADRANT_1_TOP_RIGHT_CORNER = 'quadrant_1_top_right_corner'
    QUADRANT_1_BOTTOM_RIGHT_CORNER = 'quadrant_1_bottom_right_corner'
    QUADRANT_1_BOTTOM_LEFT_CORNER = 'quadrant_1_bottom_left_corner'
    QUADRANT_2_TOP_LEFT_CORNER = 'quadrant_2_top_left_corner'
    QUADRANT_2_BOTTOM_RIGHT_CORNER = 'quadrant_2_bottom_right_corner'
    QUADRANT_2_BOTTOM_LEFT_CORNER = 'quadrant_2_bottom_left_corner'
    QUADRANT_3_TOP_RIGHT_CORNER = 'quadrant_3_top_right_corner'
    QUADRANT_3_TOP_LEFT_CORNER = 'quadrant_3_top_left_corner'
    QUADRANT_3_BOTTOM_LEFT_CORNER = 'quadrant_3_bottom_left_corner'
    QUADRANT_4_TOP_RIGHT_CORNER = 'quadrant_4_top_right_corner'
    QUADRANT_4_TOP_LEFT_CORNER = 'quadrant_4_top_left_corner'
    QUADRANT_4_BOTTOM_RIGHT_CORNER = 'quadrant_4_bottom_right_corner'

    RANDOM_INSIDE = 'random_inside'
    """
    A random position inside the screen with no pixels out of bounds.
    """
    RANDOM_OUTSIDE = 'random_outside'
    """
    A random position out of the screen. It is randomly chosen from 
    the 8 'OUT_...' options we have registered.
    """
    @classmethod
    def out_positions_as_list(cls):
        """
        Returns all the existing options out of the edges.
        """
        return [
            cls.OUT_TOP,
            cls.OUT_TOP_RIGHT,
            cls.OUT_RIGHT,
            cls.OUT_BOTTOM_RIGHT,
            cls.OUT_BOTTOM,
            cls.OUT_BOTTOM_LEFT,
            cls.OUT_LEFT,
            cls.TOP_LEFT
        ]
    
    @classmethod
    def in_positions_as_list(cls):
        """
        Returns the ScreenPosition enums that are located inside
        the screen limits.
        """
        return list(set(cls.get_all()) - set(cls.out_positions_as_list()) - set([cls.RANDOM_INSIDE]) - set([cls.RANDOM_OUTSIDE]))
    
    @classmethod
    def in_and_out_positions_as_list(cls):
        """
        Returns a list of 2 elements containing the out edge from which
        the video will come into the screen, and the opposite edge to get
        out of the screen. This has been created to animate a random slide
        transition effect. The possibilities are horizontal, diagonal and
        vertical linear sliding transitions. The first element in the list
        is the initial position and the second one, the final position. 
        """
        rnd = randrange(0, 8)
        if rnd == 0:
            positions = [ScreenPosition.OUT_RIGHT, ScreenPosition.OUT_LEFT]
        elif rnd == 1:
            positions = [ScreenPosition.OUT_TOP, ScreenPosition.OUT_BOTTOM]
        elif rnd == 2:
            positions = [ScreenPosition.OUT_BOTTOM, ScreenPosition.OUT_TOP]
        elif rnd == 3:
            positions = [ScreenPosition.OUT_TOP_LEFT, ScreenPosition.OUT_BOTTOM_RIGHT] 
        elif rnd == 4:
            positions = [ScreenPosition.OUT_TOP_RIGHT, ScreenPosition.OUT_BOTTOM_LEFT]
        elif rnd == 5:
            positions = [ScreenPosition.OUT_BOTTOM_LEFT, ScreenPosition.OUT_TOP_RIGHT]
        elif rnd == 6:
            positions = [ScreenPosition.OUT_BOTTOM_RIGHT, ScreenPosition.OUT_TOP_LEFT]
        elif rnd == 7:
            positions = [ScreenPosition.OUT_LEFT, ScreenPosition.OUT_RIGHT]

        return positions

    def get_moviepy_position(self, video, background_video):
        """
        This method will calculate the (x, y) tuple position for the provided
        'video' over the also provided 'background_video' that would be,
        hypothetically, a 1920x1080 black color background static image. The
        provided 'position' will be transformed into the (x, y) tuple according
        to our own definitions.
        """
        # TODO: Do 'video' and 'background_video' checkings
        position_tuple = None

        if self == ScreenPosition.CENTER:
            position_tuple = (get_center_x(video, background_video), get_center_y(video, background_video))

        #           Edges below
        # TOP
        elif self == ScreenPosition.OUT_TOP:
            position_tuple = ((background_video.w / 2) - (video.w / 2), -video.h)
        elif self == ScreenPosition.IN_EDGE_TOP:
            position_tuple = ((background_video.w / 2) - (video.w / 2), -(video.h / 2))
        elif self == ScreenPosition.TOP:
            position_tuple = ((background_video.w / 2) - (video.w / 2), 0)
        # TOP RIGHT
        elif self == ScreenPosition.OUT_TOP_RIGHT:
            position_tuple = (background_video.w, -video.h)
        elif self == ScreenPosition.IN_EDGE_TOP_RIGHT:
            position_tuple = (background_video.w - (video.w / 2), -(video.h / 2))
        elif self == ScreenPosition.TOP_RIGHT:
            position_tuple = (background_video.w - video.w, 0)
        # RIGHT
        elif self == ScreenPosition.OUT_RIGHT:
            position_tuple = (background_video.w, (background_video.h / 2) - (video.h / 2))
        elif self == ScreenPosition.IN_EDGE_RIGHT:
            position_tuple = (background_video.w - (video.w / 2), (background_video.h / 2) - (video.h / 2))
        elif self == ScreenPosition.RIGHT:
            position_tuple = (background_video.w - video.w, (background_video.h / 2) - (video.h / 2))
        # BOTTOM RIGHT
        elif self == ScreenPosition.OUT_BOTTOM_RIGHT:
            position_tuple = (background_video.w, background_video.h)
        elif self == ScreenPosition.IN_EDGE_BOTTOM_RIGHT:
            position_tuple = (background_video.w - (video.w / 2), background_video.h - (video.h / 2))
        elif self == ScreenPosition.BOTTOM_RIGHT:
            position_tuple = (background_video.w - video.w, background_video.h - video.h)
        # BOTTOM
        elif self == ScreenPosition.OUT_BOTTOM:
            position_tuple = ((background_video.w / 2) - (video.w / 2), background_video.h)
        elif self == ScreenPosition.IN_EDGE_BOTTOM:
            position_tuple = ((background_video.w / 2) - (video.w / 2), background_video.h - (video.h / 2))
        elif self == ScreenPosition.BOTTOM:
            position_tuple = ((background_video.w / 2) - (video.w / 2), background_video.h - video.h)
        # BOTTOM LEFT
        elif self == ScreenPosition.OUT_BOTTOM_LEFT:
            position_tuple = (-video.w, background_video.h)
        elif self == ScreenPosition.IN_EDGE_BOTTOM_LEFT:
            position_tuple = (-(video.w / 2), background_video.h - (video.h / 2))
        elif self == ScreenPosition.BOTTOM_LEFT:
            position_tuple = (0, background_video.h - video.h)
        # LEFT
        elif self == ScreenPosition.OUT_LEFT:
            position_tuple = (-video.w, (background_video.h / 2) - (video.h / 2))
        elif self == ScreenPosition.IN_EDGE_LEFT:
            position_tuple = (-(video.w / 2), (background_video.h / 2) - (video.h / 2))
        elif self == ScreenPosition.LEFT:
            position_tuple = (0, (background_video.h / 2) - (video.h / 2))
        # TOP LEFT
        elif self == ScreenPosition.OUT_TOP_LEFT:
            position_tuple = (-video.w, -video.h)
        elif self == ScreenPosition.IN_EDGE_TOP_LEFT:
            position_tuple = (-(video.w / 2), -(video.h / 2))
        elif self == ScreenPosition.TOP_LEFT:
            position_tuple = (0, 0)

        # HALF POSITIONS
        elif self == ScreenPosition.HALF_TOP:
            position_tuple = (background_video.w / 2 - video.w / 2, background_video.h / 4 - video.h / 2)
        elif self == ScreenPosition.HALF_RIGHT:
            position_tuple = (3 * background_video.w / 4 - video.w / 2, background_video.h / 2 - video.h / 2)
        elif self == ScreenPosition.HALF_BOTTOM:
            position_tuple = (background_video.w / 2 - video.w / 2, 3 * background_video.h / 4 - video.h / 2)
        elif self == ScreenPosition.HALF_LEFT:
            position_tuple = (background_video.w / 4 - video.w / 2, background_video.h / 2 - video.h / 2)
        elif self == ScreenPosition.HALF_TOP_RIGHT:
            position_tuple = (3 * background_video.w / 4 - video.w / 2, background_video.h / 4 - video.h / 2)
        elif self == ScreenPosition.HALF_BOTTOM_RIGHT:
            position_tuple = (3 * background_video.w / 4 - video.w / 2, 3 * background_video.h / 4 - video.h / 2)
        elif self == ScreenPosition.HALF_BOTTOM_LEFT:
            position_tuple = (background_video.w / 4 - video.w / 2, 3 * background_video.h / 4 - video.h / 2)
        elif self == ScreenPosition.HALF_TOP_LEFT:
            position_tuple = (background_video.w / 4 - video.w / 2, background_video.h / 4 - video.h / 2)

        # QUADRANT CORNERS
        elif self == ScreenPosition.QUADRANT_1_TOP_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, 0)
        elif self == ScreenPosition.QUADRANT_1_BOTTOM_LEFT_CORNER:
            position_tuple = (0, background_video.h / 2 - video.h)
        elif self == ScreenPosition.QUADRANT_1_BOTTOM_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, background_video.h / 2 - video.h)
        elif self == ScreenPosition.QUADRANT_2_BOTTOM_LEFT_CORNER:
            position_tuple = (background_video.w / 2, background_video.h / 2 - video.h)
        elif self == ScreenPosition.QUADRANT_2_BOTTOM_RIGHT_CORNER:
            position_tuple = (background_video.w - video.w, background_video.h / 2 - video.h)
        elif self == ScreenPosition.QUADRANT_2_TOP_LEFT_CORNER:
            position_tuple = (background_video.w / 2, 0)
        elif self == ScreenPosition.QUADRANT_3_BOTTOM_LEFT_CORNER:
            position_tuple = (background_video.w / 2, background_video.h - video.h)
        elif self == ScreenPosition.QUADRANT_3_TOP_LEFT_CORNER:
            position_tuple = (background_video.w / 2, background_video.h / 2)
        elif self == ScreenPosition.QUADRANT_3_TOP_RIGHT_CORNER:
            position_tuple = (background_video.w - video.w, background_video.h / 2)
        elif self == ScreenPosition.QUADRANT_4_BOTTOM_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, background_video.h - video.h)
        elif self == ScreenPosition.QUADRANT_4_TOP_LEFT_CORNER:
            position_tuple = (0, background_video.h / 2)
        elif self == ScreenPosition.QUADRANT_4_TOP_RIGHT_CORNER:
            position_tuple = (background_video.w / 2 - video.w, background_video.h / 2)

        # RANDOMs
        elif self == ScreenPosition.RANDOM_INSIDE:
            lower_limit = ScreenPosition.TOP_LEFT.get_moviepy_position(video, background_video)
            upper_limit = ScreenPosition.BOTTOM_RIGHT.get_moviepy_position(video, background_video)
            position_tuple = (randrange(lower_limit[0], upper_limit[0]), randrange(lower_limit[1], upper_limit[1]))
        elif self == ScreenPosition.RANDOM_OUTSIDE:
            # By now I'm choosing one of the 'OUT' available options
            position_tuple = randomchoice(ScreenPosition.out_positions_as_list())

        return position_tuple


# Effects
SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/17rw299LIMTEciS7WDgFbIEZ7DckpHsC7/view?usp=sharing'
PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/1xr6x_yeU9SF5guB6KA9cLqbmVUwxBMi8/view?usp=sharing'

# Manim
WHOOSH_TITLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/1WPS8uWB1LTuzPzxQ2Zcp1FpwvLM3fhM5/view?usp=sharing'
MINECRAFT_ITEM_DROP_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/1Dzeb6Qae4UdpmuA6U9d6t3MXnvhzukzg/view?usp=sharing'


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.parser import VideoParser
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union


def apply_effect_to_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], effect: VideoEffect, **kwargs):
    """
    This method will apply the provided 'effect' (if valid) to the also
    provided whole 'video' and will return the new clip with the effect
    applied.

    You can pass any effect customization parameter through the 'kwargs'
    parameters, but I recommend you to avoid this method and instantiate
    a valid 'effect', passing the 'video' to the initializer and using 
    the '.apply()' method like in this line below:
    > BlurMoviepyEffect(video).apply()

    The text above is easier for the developer because can check the 
    parameters accepted by the effect in the initializing process.
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if not effect:
        raise Exception('No "effect" provided.')
    
    # We check if the parent of the effect is one of the 
    # expected ones
    # TODO: Maybe issubclass(effect, VideoEffect) (?)
    effect_parent_classes = effect.__bases__
    if VideoEffect in effect_parent_classes:
        raise Exception('Provided "effect" is not valid.')
    
    video = VideoParser.to_moviepy(video)
    
    # TODO: This is untested with the new effect version
    return effect.apply(video, **kwargs)

"""
# The effects that we are using here are the ones prepared for the
# segment, with information about the time and that stuff.

# Ok, this is a general method, not our main flow method.
# We just need to apply the effect in the video, so lets
# check that the effect is valid, apply on the video if
# possible, and return the video.
"""


MANIM_RESOURCES_FOLDER = 'resources/manim/'

# Manim core
MANDATORY_CONFIG_PARAMETER = 1
OPTIONAL_CONFIG_PARAMETER = 0
# I obtained manim dimensions from here: https://docs.manim.community/en/stable/faq/general.html#what-are-the-default-measurements-for-manim-s-scene
SCENE_HEIGHT = 8 # 1080 equivalent
"""
8 (manim dimension). It is the equivalent, for our suposed scene,
to 1080 in pixels.
"""
SCENE_WIDTH = 14 + (2 / 9) # 1920 equivalent
"""
14 + 2/9 (manim dimension). It is the equivalent, for our suposed
scene, to 1920 in pixels.
"""
HALF_SCENE_HEIGHT = SCENE_HEIGHT / 2
"""
8 / 2 = 4 (manim dimension)
"""
HALF_SCENE_WIDTH = SCENE_WIDTH / 2
"""
(14 + 2/9) / 2 = 7 + 1/9 (manim dimension)
"""
LEFT_MARGIN = -HALF_SCENE_WIDTH
UP_MARGIN = HALF_SCENE_HEIGHT
STANDARD_HEIGHT = 1080
"""
This is our standard height (in pixels) as we will create animation
videos of 1920x1080.
"""
STANDARD_WIDTH = 1920
"""
This is our standard width (in pixels) as we will create animation
videos of 1920x1080.
"""




"""
Nice help: https://www.bannerbear.com/blog/how-to-use-ffmpeg-in-python-with-examples/
Official doc: https://www.ffmpeg.org/ffmpeg-resampler.html
More help: https://kkroening.github.io/ffmpeg-python/
Nice guide: https://img.ly/blog/ultimate-guide-to-ffmpeg/
Available flags: https://gist.github.com/tayvano/6e2d456a9897f55025e25035478a3a50

Interesting usage: https://stackoverflow.com/a/20325676
Maybe avoid writting on disk?: https://github.com/kkroening/ffmpeg-python/issues/500#issuecomment-792281072
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.audio.parser import AudioParser
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.image.parser import ImageParser
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.file.writer import write_file
from typing import Union
from subprocess import run


class FfmpegAudioCodec(Enum):
    """
    TODO: Fill this

    Should be used in the **-c:a {codec}** flag.
    """
    AAC = 'aac'
    """
    Default encoder.
    """
    AC3 = 'ac3'
    """
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-ac3-and-ac3_005ffixed
    """
    AC3_FIXED = 'ac3_fixed'
    """
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-ac3-and-ac3_005ffixed
    """
    FLAC = 'flac'
    """
    FLAC (Free Lossless Audio Codec) Encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-flac-2
    """
    OPUS = 'opus'
    """
    This is a native FFmpeg encoder for the Opus format. Currently, it’s
    in development and only implements the CELT part of the codec. Its
    quality is usually worse and at best is equal to the libopus encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-opus
    """
    LIBFDK_AAC = 'libfdk_aac'
    """
    libfdk-aac AAC (Advanced Audio Coding) encoder wrapper. The libfdk-aac
    library is based on the Fraunhofer FDK AAC code from the Android project.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libfdk_005faac
    """
    LIBLC3 = 'liblc3'
    """
    liblc3 LC3 (Low Complexity Communication Codec) encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-liblc3
    """
    LIBMP3LAME = 'libmp3lame'
    """
    LAME (Lame Ain’t an MP3 Encoder) MP3 encoder wrapper.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libmp3lame-1
    """
    LIBOPENCORE_AMRNB = 'libopencore_amrnb'
    """
    OpenCORE Adaptive Multi-Rate Narrowband encoder.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libopencore_002damrnb-1ss
    """
    LIBOPUS = 'libopus'
    """
    libopus Opus Interactive Audio Codec encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libopus-1
    """
    LIBSHINE = 'libshine'
    """
    Shine Fixed-Point MP3 encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libshine-1
    """
    LIBTWOLAME = 'libtwolame'
    """
    TwoLAME MP2 encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libtwolame
    """
    LIBVO_AMRWBENC = 'libvo-amrwbenc'
    """
    VisualOn Adaptive Multi-Rate Wideband encoder.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libvo_002damrwbenc
    """
    LIBVORBIS = 'libvorbis'
    """
    libvorbis encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libvorbis
    """
    MJPEG = 'mjpeg'
    """
    Motion JPEG encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-mjpeg
    """
    WAVPACK = 'wavpack'
    """
    WavPack lossless audio encoder.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-wavpack
    """

class FfmpegVideoCodec(Enum):
    """
    These are the video codecs available as Enums. The amount of codecs
    available depends on the ffmpeg built version.
    
    Should be used in the **-c:v {codec}** flag.
    """
    A64_MULTI = 'a64_multi'
    """
    A64 / Commodore 64 multicolor charset encoder. a64_multi5 is extended with 5th color (colram).

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-a64_005fmulti_002c-a64_005fmulti5
    """
    A64_MULTI5 = 'a64_multi5'
    """
    A64 / Commodore 64 multicolor charset encoder. a64_multi5 is extended with 5th color (colram).

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-a64_005fmulti_002c-a64_005fmulti5
    """
    CINEPAK = 'Cinepak'
    """
    Cinepak aka CVID encoder. Compatible with Windows 3.1 and vintage MacOS.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-Cinepak
    """
    GIF = 'GIF'
    """
    GIF image/animation encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-Cinepak
    """
    HAP = 'Hap'
    """
    Vidvox Hap video encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-Hap
    """
    JPEG2000 = 'jpeg2000'
    """
    The native jpeg 2000 encoder is lossy by default

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-jpeg2000
    """
    LIBRAV1E = 'librav1e'
    """
    rav1e AV1 encoder wrapper.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-librav1e
    """
    LIBAOM_AV1 = 'libaom-av1'
    """
    libaom AV1 encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libaom_002dav1
    """
    # TODO: Continue with this (https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libsvtav1)
    QTRLE = 'qtrle'
    """
    TODO: Find information about this video codec.

    More info: ???
    """
    PRORES = 'prores'
    """
    Apple ProRes encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-ProRes
    """

class FfmpegVideoFormat(Enum):
    """
    Enum list to simplify the way we choose a video format for
    the ffmpeg command. This should be used with the FfmpegFlag
    '-f' flag that forces that video format.

    Should be used in the **-f {format}** flag.
    """
    CONCAT = 'concat'
    """
    The format will be the concatenation.
    """
    AVI = 'avi'
    """
    Avi format.

    # TODO: Explain more
    """
    PNG = 'png'
    """
    # TODO: Look for mor information about this vcodec
    # TODO: I don't know if this one is actually an FfmpegVideoFormat
    # or if I need to create another Enum class. This option us used
    # in the '-vcodec' option, and the other ones are used in the
    # 'c:v' option.
    """
    # TODO: Keep going

class FfmpegFilter(Enum):
    """
    Enum list to simplify the way we use a filter for the
    ffmpeg command.

    Should be used in the **-filter {filter}** flag.
    """
    THUMBNAIL = 'thumbnail'
    """
    Chooses the most representative frame of the video to be used
    as a thumbnail.
    """

class FfmpegPixelFormat(Enum):
    """
    Enum list to simplify the way we use a pixel format for
    the ffmpeg command.

    Should be used in the **-pix_fmt {format}** flag.
    """
    YUV420p = 'yuv420p'
    """
    This is de default value. TODO: Look for more information about it
    """
    RGB24 = 'rgb24'
    """
    TODO: Look for more information about this pixel format.
    """
    ARGB = 'argb'
    """
    TODO: Look for more information about this pixel format
    """
    YUVA444P10LE = 'yuva444p10le'
    """
    TODO: Look for more information about this pixel format
    """

class FfmpegFlag:
    """
    Class to simplify the way we push flags into the ffmpeg command.
    """
    overwrite: str = '-y'
    """
    Overwrite the output file if existing.

    Notation: **-y**
    """

    @classmethod
    def force_format(cls, format: FfmpegVideoFormat):
        """
        Force the output format to be the provided 'format'.

        Notation: **-f {format}**
        """
        format = FfmpegVideoFormat.to_enum(format).value

        return f'-f {format}'
    
    @classmethod
    def safe_routes(cls, value: int):
        """
        To enable or disable unsafe paths.

        Notation: **-safe {value}**
        """
        # TODO: Check that 'value' is a number between -1 and 1

        return f'-safe {str(value)}'
    
    @classmethod
    def input(cls, input: str):
        """
        To set the input (or inputs) we want.

        Notation: **-i {input}**
        """
        # TODO: I don't know how to check this or format it from 'input' param

        return f'-i {input}'
    
    @classmethod
    def audio_codec(cls, codec: Union[FfmpegAudioCodec, str]):
        """
        Sets the general audio codec.

        Notation: **-c:a {codec}**
        """
        # We cannot control the big amount of audio codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegAudioCodec.to_enum(codec).value
        except:
            pass

        return f'-c:a {codec}'
    
    @classmethod
    def video_codec(cls, codec: Union[FfmpegVideoCodec, str]):
        """
        Sets the general video codec.

        Notation: **-c:v {codec}**
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegVideoCodec.to_enum(codec).value
        except:
            pass

        return f'-c:v {codec}'

    @classmethod
    def v_codec(cls, codec: Union[FfmpegVideoCodec, str]):
        """
        Sets the video codec.

        TODO: I don't know exactly the difference between '-c:v {codec}'
        and the '-vcodec' generated in this method. I keep this method
        until I actually find the difference. I don't even know if the
        video codecs I can provide as values are the same as in the other
        method.

        Notation: **-vcodec {codec}**
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegVideoCodec.to_enum(codec).value
        except:
            pass

        return f'-vcodec {codec}'

    @classmethod
    def codec(cls, codec: Union[FfmpegVideoCodec, FfmpegAudioCodec, str]):
        """
        Sets the general codec with '-c {codec}'.

        -c copy indica que se deben copiar los flujos de audio y video sin recodificación, lo que hace que la operación sea rápida y sin pérdida de calidad. TODO: Turn this 'copy' to AudioCodec and VideoCodec (?)

        Notation: **-c {codec}**
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter

        # TODO: Validate provided 'codec'
        # TODO: This method has a variation, it can be '-c:a' or '-c:v'
        if not isinstance(codec, (FfmpegVideoCodec, FfmpegAudioCodec)):
            try:
                codec = FfmpegVideoCodec.to_enum(codec)
            except:
                pass

        if not isinstance(codec, (FfmpegVideoCodec, FfmpegAudioCodec)):
            try:
                codec = FfmpegAudioCodec.to_enum(codec)
            except:
                pass

        if isinstance(codec, (FfmpegVideoCodec, FfmpegAudioCodec)):
            codec = codec.value

        return f'-c {codec}'
    
    @classmethod
    def map(cls, map: str):
        """
        Set input stream mapping.
        -map [-]input_file_id[:stream_specifier][,sync_file_id[:stream_s set input stream mapping

        # TODO: Improve this

        Notation: **-map {map}**
        """
        return f'-map {map}'
    
    @classmethod
    def filter(cls, filter: FfmpegFilter):
        """
        Sets the expected filter to be used.

        Notation: **-filter {filter}**
        """
        filter = FfmpegFilter.to_enum(filter).value

        return f'-filter {filter}'
    
    @classmethod
    def frame_rate(cls, frame_rate: int):
        """
        Sets the frame rate (Hz value, fraction or abbreviation)

        Notation: **-r {frame_rate}**
        """
        # TODO: Validate 'frame_rate'

        return f'-r {str(frame_rate)}'
    
    @classmethod
    def pixel_format(cls, format: FfmpegPixelFormat):
        """
        Sets the pixel format.

        Notation: **-pix_fmt {format}**
        """
        format = FfmpegPixelFormat.to_enum(format).value

        return f'-pix_fmt {format}'


class FfmpegHandler:
    """
    Class to simplify and encapsulate ffmpeg functionality.
    """
    @classmethod
    def _validate_video_filename(cls, video_filename: str):
        # TODO: Validate and raise Exception if invalid
        pass

    @classmethod
    def write_concat_file(cls, filenames: str):
        """
        Writes the files to concat in a temporary text file with
        the required format and returns that file filename. This
        is required to use different files as input.
        """
        text = ''
        for filename in filenames:
            text += f"file '{filename}'\n"

        # TODO: Maybe this below is interesting for the 'yta_general_utils.file.writer'
        # open('concat.txt', 'w').writelines([('file %s\n' % input_path) for input_path in input_paths])
        filename = create_temp_filename('concat_ffmpeg.txt')
        write_file(text, filename)

        return filename

    @classmethod
    def run_command(cls, args: list[str]):
        """
        Runs an ffmpeg provided with the provided 'args' as a subprocess.
        """
        # TODO: Clean args (?)
        # Remove 'None' args, our logic allow that to make it easier
        args = [arg for arg in args if arg is not None]

        # TODO: Validate args

        run(f"ffmpeg {' '.join(args)}")

    # TODO: Check this one below
    @classmethod
    def get_audio_from_video_deprecated(cls, video_filename: str, codec: FfmpegAudioCodec = None, output_filename: str = None):
        """
        Exports the audio of the provided video to a local file named
        'output_filename' if provided or as a temporary file.

        # TODO: This has not been tested yet.

        This methods returns a tuple with the audio as a moviepy audio 
        in the first place and the filename of the file generated in
        the second place.
        """
        cls._validate_video_filename(video_filename)
        
        if not output_filename:
            output_filename = create_temp_filename('temp_audio.mp3')
        # TODO: Validate 'output_filename' is a valid audio filename

        if codec:
            codec = FfmpegAudioCodec.to_enum(codec)

        cls.run_command([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.audio_codec(codec) if codec else None,
            output_filename
        ])

        return AudioParser.to_audiofileclip(output_filename), output_filename
    
    @classmethod
    def get_audio_from_video(cls, video_filename: str, output_filename: str = None):
        """
        Exports the audio of the provided video to a local file named
        'output_filename' if provided or as a temporary file.

        This methods returns a tuple with the audio as a moviepy audio 
        in the first place and the filename of the file generated in
        the second place.
        """
        cls._validate_video_filename(video_filename)
        
        if not output_filename:
            output_filename = create_temp_filename('temp_audio.mp3')

        # TODO: Verify valid output_filename extension

        cls.run_command([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.map('0:1'),
            output_filename
        ])

        return AudioParser.to_audiofileclip(output_filename), output_filename

    @classmethod
    def get_best_thumbnail(cls, video_filename: str, output_filename: str = None):
        """
        Gets the best thumbnail of the provided 'video_filename'.

        This methods returns a tuple with the thumbnail as a pillow 
        image in the first place and the filename of the file generated
        in the second place.
        """
        cls._validate_video_filename(video_filename)

        if not output_filename:
            output_filename = create_temp_filename('temp_thumbnail.png')
        # TODO: Verify valid extension

        cls.run_command([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.filter(FfmpegFilter.THUMBNAIL),
            output_filename
        ])

        return ImageParser.to_pillow(output_filename), output_filename
    
    @classmethod
    def concatenate_videos(cls, video_filenames: str, output_filename: str = None):
        """
        Concatenates the provided 'video_filenames' in the order in
        which they are provided.

        This methods returns a tuple with the new video as a moviepy
        video in the first place and the filename of the file generated
        in the second place.
        """
        for video_filename in video_filenames:
            cls._validate_video_filename(video_filename)

        if not output_filename:
            output_filename = create_temp_filename('concatenated_video.mp4')

        concat_filename = cls.write_concat_file(video_filenames)

        cls.run_command([
            FfmpegFlag.overwrite,
            FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
            FfmpegFlag.safe_routes(0),
            FfmpegFlag.input(concat_filename),
            FfmpegFlag.codec('copy'),
            output_filename
        ])

        return VideoParser.to_moviepy(output_filename), output_filename
    
    @classmethod
    def concatenate_images(cls, image_filenames: str, frame_rate = 60, pixel_format: FfmpegPixelFormat = FfmpegPixelFormat.YUV420p, output_filename: str = None):
        """
        Concatenates the provided 'image_filenames' in the order in
        which they are provided.

        This methods returns a tuple with the new video as a moviepy
        video in the first place and the filename of the file generated
        in the second place.
        """
        for image_filename in image_filenames:
            cls._validate_video_filename(image_filename)

        if not output_filename:
            output_filename = create_temp_filename('concatenated_video.mp4')

        concat_filename = cls.write_concat_file(image_filenames)

        cls.run_command([
            FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
            FfmpegFlag.safe_routes(0),
            FfmpegFlag.overwrite,
            FfmpegFlag.frame_rate(frame_rate),
            FfmpegFlag.input(concat_filename),
            FfmpegFlag.video_codec(FfmpegVideoCodec.QTRLE),
            FfmpegFlag.pixel_format(pixel_format), # I used 'argb' in the past
            output_filename
        ])

        return VideoParser.to_moviepy(output_filename), output_filename

    # TODO: Keep going

    # https://www.reddit.com/r/ffmpeg/comments/ks8zfs/comment/gieu7x6/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
    # https://stackoverflow.com/questions/38368105/ffmpeg-custom-sequence-input-images/51618079#51618079
    # https://stackoverflow.com/a/66014158


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.resources.video.effect.sound.drive_urls import PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.video.edition.effect.moviepy.blink_video_effect import BlinkVideoEffect
from yta_multimedia.resources import Resource
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, CompositeAudioClip, AudioFileClip, ColorClip, VideoClip
from typing import Union


class PhotoVideoEffect(VideoEffect):
    """
    Simulates that a photo is taken by making a white blink and
    a camera click sound. This effect doesn't freeze the video,
    it is just a white blink with a camera photo sound being
    played.
    """
    
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        Applies the effect to the provided 'clip'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        TMP_FILENAME = Resource.get(PHOTO_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'sounds/photo_taken.mp3')

        # We force the effect to be last as much as the clip
        video = BlinkVideoEffect.apply(video, [255, 255, 255])

        effect_duration = 0.2
        if video.duration < effect_duration:
            effect_duration = video.duration

        video.audio = CompositeAudioClip([
            video.audio,
            AudioFileClip(TMP_FILENAME).set_duration(effect_duration)
        ])

        return video



from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames import VideoFrameExtractor
from yta_multimedia.video.generation import generate_video_from_image
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, concatenate_videoclips
from typing import Union


# TODO: Move this to a better place
class ExtendVideoMode(Enum):
    """
    This is a Enum to set the parameter option to extend the video
    duration with one of these modes (strategies).
    """
    LOOP = 'loop'
    """
    This mode will make the video loop (restart from the begining)
    until it reaches the expected duration.
    """
    FREEZE_LAST_FRAME = 'freeze_last_frame'
    """
    This mode will freeze the last frame of the video and extend 
    it until it reaches the expected duration.
    """


def set_video_duration(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], duration = float, mode: ExtendVideoMode = ExtendVideoMode.LOOP):
    """
    This method will return a copy of the provided 'video' with the desired
    'duration' by applying crops or loops. If the provided 'duration' is
    lower than the actual 'video' duration, it will be shortened. If it is
    greater, it will be looped until we reach the desired 'duration'.

    The 'mode' provided will determine the way in which we extend the video
    duration if needed.

    This method makes a 'video.copy()' internally to work and avoid problems.
    """
    video = VideoParser.to_moviepy(video)

    if not duration:
        raise Exception('No "duration" provided.')
    
    if not mode:
        mode = ExtendVideoMode.LOOP

    mode = ExtendVideoMode.to_enum(mode)

    final_video = video.copy()

    if video.duration > duration:
        final_video = final_video.subclip(0, duration)
    elif video.duration < duration:
        if mode == ExtendVideoMode.LOOP:
            times_to_loop = (int) (duration / video.duration) - 1
            remaining_time = duration % video.duration
            for _ in range(times_to_loop):
                final_video = concatenate_videoclips([final_video, video])
            final_video = concatenate_videoclips([final_video, video.subclip(0, remaining_time)])
        elif mode == ExtendVideoMode.FREEZE_LAST_FRAME:
            remaining_time = duration - video.duration
            frame = VideoFrameExtractor.get_frame_by_time(video, video.duration)
            frame_freezed_video = generate_video_from_image(frame, remaining_time)
            final_video = concatenate_videoclips([video, frame_freezed_video])

    return final_video


from yta_multimedia.video.generation.manim.constants import HALF_SCENE_HEIGHT, HALF_SCENE_WIDTH
from yta_multimedia.video.generation.manim.utils.dimensions import width_to_manim_width, height_to_manim_height
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.random import randrangefloat
from random import random, choice as randchoice


class ManimScreenPosition(Enum):
    """
    Enum class that represents a position inside the screen
    scene. This can be used to position a video in one of
    these different regions.
    """
    OUT_TOP_LEFT = 'out_top_left'
    """
    Out of the screen, on the top left corner, just one pixel
    out of bounds.
    """
    IN_EDGE_TOP_LEFT = 'in_edge_top_left'
    """
    The center of the video is on the top left corner, so only
    the bottom right quarter part of the video is shown (inside
    the screen).
    """
    TOP_LEFT = 'top_left'
    """
    The video is completely visible, just at the top left 
    corner of the screen.
    """
    OUT_TOP = 'out_top'
    IN_EDGE_TOP = 'in_edge_top'
    TOP = 'top'
    OUT_TOP_RIGHT = 'out_top_right'
    IN_EDGE_TOP_RIGHT = 'in_edge_top_right'
    TOP_RIGHT = 'top_right'
    CENTER = 'center'
    OUT_RIGHT = 'out_right'
    IN_EDGE_RIGHT = 'in_edge_right'
    RIGHT = 'right'
    OUT_BOTTOM_RIGHT = 'out_bottom_right'
    IN_EDGE_BOTTOM_RIGHT = 'in_edge_bottom_right'
    BOTTOM_RIGHT = 'bottom_right'
    OUT_BOTTOM = 'out_bottom'
    IN_EDGE_BOTTOM = 'in_edge_bottom'
    BOTTOM = 'bottom'
    OUT_BOTTOM_LEFT = 'out_bottom_left'
    IN_EDGE_BOTTOM_LEFT = 'in_edge_bottom_left'
    BOTTOM_LEFT = 'bottom_left'
    OUT_LEFT = 'out_left'
    IN_EDGE_LEFT = 'in_edge_left'
    LEFT = 'left'

    HALF_TOP = 'half_top'
    HALF_TOP_RIGHT = 'half_top_right'
    HALF_RIGHT = 'half_right'
    HALF_BOTTOM_RIGHT = 'half_bottom_right'
    HALF_BOTTOM = 'half_bottom'
    HALF_BOTTOM_LEFT = 'half_bottom_left'
    HALF_LEFT = 'half_left'
    HALF_TOP_LEFT = 'half_top_left'

    QUADRANT_1_TOP_RIGHT_CORNER = 'quadrant_1_top_right_corner'
    QUADRANT_1_BOTTOM_RIGHT_CORNER = 'quadrant_1_bottom_right_corner'
    QUADRANT_1_BOTTOM_LEFT_CORNER = 'quadrant_1_bottom_left_corner'
    QUADRANT_2_TOP_LEFT_CORNER = 'quadrant_2_top_left_corner'
    QUADRANT_2_BOTTOM_RIGHT_CORNER = 'quadrant_2_bottom_right_corner'
    QUADRANT_2_BOTTOM_LEFT_CORNER = 'quadrant_2_bottom_left_corner'
    QUADRANT_3_TOP_RIGHT_CORNER = 'quadrant_3_top_right_corner'
    QUADRANT_3_TOP_LEFT_CORNER = 'quadrant_3_top_left_corner'
    QUADRANT_3_BOTTOM_LEFT_CORNER = 'quadrant_3_bottom_left_corner'
    QUADRANT_4_TOP_RIGHT_CORNER = 'quadrant_4_top_right_corner'
    QUADRANT_4_TOP_LEFT_CORNER = 'quadrant_4_top_left_corner'
    QUADRANT_4_BOTTOM_RIGHT_CORNER = 'quadrant_4_bottom_right_corner'

    RANDOM_INSIDE = 'random_inside'
    """
    A random position inside the screen with no pixels out of bounds.
    It is randomly chosen from one of all the options inside the limits
    we have.
    """
    RANDOM_OUTSIDE = 'random_outside'
    """
    A random position out of the screen limits. It is randomly chosen 
    from one of all the options outside the limits we have.
    """
    # TODO: Add more positions

    @classmethod
    def out_positions_as_list(cls):
        """
        Returns the ManimScreenPosition enums that are located out of
        the screen limits.
        """
        return [
            cls.OUT_TOP_LEFT,
            cls.OUT_TOP,
            cls.OUT_RIGHT,
            cls.OUT_BOTTOM_RIGHT,
            cls.OUT_BOTTOM,
            cls.OUT_BOTTOM_LEFT,
            cls.OUT_LEFT
        ]
    
    @classmethod
    def in_positions_as_list(cls):
        """
        Returns the ManimScreenPosition enums that are located inside
        the screen limits.
        """
        return list(set(cls.get_all()) - set(cls.out_positions_as_list()) - set([cls.RANDOM_INSIDE]) - set([cls.RANDOM_OUTSIDE]))

    def get_limits(self):
        """
        Return the left, right, top and bottom limits for this
        screen position. This edges represent the limits of the
        region in which the video should be placed to fit this
        screen position.

        We consider each screen region as a limited region of
        half of the scene width and height.

        Corner limits:
        [-7-1/9,  4, 0]   [0,  4, 0]   [7+1/9,  4, 0]
        [-7-1/9,  0, 0]   [0,  0, 0]   [7+1/9,  0, 0]
        [-7-1/9, -4, 0]   [0, -4, 0]   [7+1/9, -4, 0]
        """
        # TODO: I think I should consider regions of 1/8 of width and height
        # so 1 quadrant is divided into 4 pieces and I build all the positions
        # for also those quadrants
        # TODO: I'm missing the QUADRANT_1_HALF_TOP and HALF_TOP_OUT, ...
        if self == ManimScreenPosition.TOP:
            return -HALF_SCENE_WIDTH / 2, HALF_SCENE_WIDTH / 2, HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.BOTTOM:
            return -HALF_SCENE_WIDTH / 2, HALF_SCENE_WIDTH / 2, -HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.LEFT:
            return -HALF_SCENE_WIDTH, 0, HALF_SCENE_HEIGHT / 2, -HALF_SCENE_HEIGHT / 2
        elif self == ManimScreenPosition.RIGHT:
            return 0, HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT / 2, -HALF_SCENE_HEIGHT / 2
        # TODO: Add missing

    def get_position(self, width: float, height: float):
        """
        Return the position in which the mobject must be placed to
        be exactly in this position.
        """
        # TODO: 'width' and 'height' must be manim

        if self == ManimScreenPosition.TOP:
            return 0, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.TOP_RIGHT:
            return HALF_SCENE_WIDTH - width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.RIGHT:
            return HALF_SCENE_WIDTH - width / 2, 0, 0
        elif self == ManimScreenPosition.BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH - width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.BOTTOM:
            return 0, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH + width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.LEFT:
            return -HALF_SCENE_WIDTH + width / 2, 0, 0
        elif self == ManimScreenPosition.TOP_LEFT:
            return -HALF_SCENE_WIDTH + width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.CENTER:
            return 0, 0, 0
        elif self == ManimScreenPosition.IN_EDGE_TOP_LEFT:
            return -HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.OUT_TOP_LEFT:
            return -HALF_SCENE_WIDTH - width / 2, HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.IN_EDGE_TOP:
            return 0, HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.OUT_TOP:
            return 0, HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.IN_EDGE_TOP_RIGHT:
            return HALF_SCENE_WIDTH, HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.OUT_TOP_RIGHT:
            return HALF_SCENE_WIDTH + width / 2, HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.IN_EDGE_RIGHT:
            return HALF_SCENE_WIDTH, 0, 0
        elif self == ManimScreenPosition.OUT_RIGHT:
            return HALF_SCENE_WIDTH + width / 2, 0, 0
        elif self == ManimScreenPosition.IN_EDGE_BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH, -HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.OUT_BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH + width / 2, -HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.IN_EDGE_BOTTOM:
            return 0, -HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.OUT_BOTTOM:
            return 0, -HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.IN_EDGE_BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH, -HALF_SCENE_HEIGHT, 0
        elif self == ManimScreenPosition.OUT_BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH - width / 2, -HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.IN_EDGE_LEFT:
            return -HALF_SCENE_WIDTH, 0, 0
        elif self == ManimScreenPosition.OUT_LEFT:
            return -HALF_SCENE_WIDTH - width / 2, 0, 0
        elif self == ManimScreenPosition.HALF_TOP:
            return 0, HALF_SCENE_HEIGHT / 2, 0
        elif self == ManimScreenPosition.HALF_TOP_RIGHT:
            return HALF_SCENE_WIDTH / 2, HALF_SCENE_HEIGHT / 2, 0
        elif self == ManimScreenPosition.HALF_RIGHT:
            return HALF_SCENE_WIDTH / 2, 0, 0
        elif self == ManimScreenPosition.HALF_BOTTOM_RIGHT:
            return HALF_SCENE_WIDTH / 2, -HALF_SCENE_HEIGHT / 2, 0
        elif self == ManimScreenPosition.HALF_BOTTOM:
            return 0, -HALF_SCENE_HEIGHT / 2, 0
        elif self == ManimScreenPosition.HALF_BOTTOM_LEFT:
            return -HALF_SCENE_WIDTH / 2, -HALF_SCENE_HEIGHT / 2, 0
        elif self == ManimScreenPosition.HALF_LEFT:
            return -HALF_SCENE_WIDTH / 2, 0, 0
        elif self == ManimScreenPosition.HALF_TOP_LEFT:
            return -HALF_SCENE_WIDTH / 2, HALF_SCENE_HEIGHT / 2, 0
        elif self == ManimScreenPosition.QUADRANT_1_TOP_RIGHT_CORNER:
            return -width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_1_BOTTOM_RIGHT_CORNER:
            return -width / 2, height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_1_BOTTOM_LEFT_CORNER:
            return -HALF_SCENE_WIDTH + width / 2, height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_2_TOP_LEFT_CORNER:
            return width / 2, HALF_SCENE_HEIGHT - height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_2_BOTTOM_RIGHT_CORNER:
            return HALF_SCENE_WIDTH - width / 2, height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_2_BOTTOM_LEFT_CORNER:
            return width / 2, height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_3_TOP_LEFT_CORNER:
            return width / 2, -height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_3_TOP_RIGHT_CORNER:
            return HALF_SCENE_WIDTH - width / 2, -height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_3_BOTTOM_LEFT_CORNER:
            return width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_4_TOP_LEFT_CORNER:
            return -HALF_SCENE_WIDTH + width / 2, -height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_4_TOP_RIGHT_CORNER:
            return -width / 2, -height / 2, 0
        elif self == ManimScreenPosition.QUADRANT_4_BOTTOM_RIGHT_CORNER:
            return -width / 2, -HALF_SCENE_HEIGHT + height / 2, 0
        elif self == ManimScreenPosition.RANDOM_INSIDE:
            return randchoice(ManimScreenPosition.in_positions_as_list()).get_position(width, height)
        elif self == ManimScreenPosition.RANDOM_OUTSIDE:
            return randchoice(ManimScreenPosition.out_positions_as_list()).get_position(width, height)
        
    def get_random_position(self, width: float, height: float):
        """
        Calculate the position in which the center of the element 
        with the provided 'width' and 'height' must be placed to
        obtain this position.

        The position will be provided as a 3D vector but the z
        axis will be always 0.

        The provided 'width' and 'height' must be in pixels.
        """
        # TODO: Check if the provided 'width' and 'height' are in
        # in pixels and valid or if not

        # TODO: By now I'm just using the limits as numeric limits
        # for random position that will be used as the center of the
        # video, but we will need to consider the video dimensions 
        # in a near future to actually position it well, because the
        # video can be out of the scene right now with this approach
        left, right, top, bottom = self.get_limits()

        x, y = randrangefloat(left, right, width_to_manim_width(1)), randrangefloat(top, bottom, height_to_manim_height(1))

        # If video is larger than HALF/2 it won't fit correctly.
        if width > HALF_SCENE_WIDTH or height > HALF_SCENE_HEIGHT:
            # TODO: Video is bigger than the region, we cannot make
            # it fit so... what can we do (?)
            return x, y

        if x - width / 2 < left:
            x += left - (x - width / 2)
        if x + width / 2 > right:
            x -= (x + width / 2) - right
        if y - height / 2 < bottom:
            y += bottom - (y - height / 2)
        if y + height / 2 > top:
            y -= (y + height / 2) - top

        return x, y, 0
        

class ManimPosition:
    @staticmethod
    def min_and_max_x(width: float):
        """
        Calculates the minimum and maximum possible 'x' to make the 
        object with the provided 'width' fit the manim screen, that
        means to be inside of it and appear in the video.

        The provided 'width' must be a manim width, not a width in
        pixels.

        This method returns the 'min, max' pair of minimum and
        maximum possible x values.
        """
        return -HALF_SCENE_WIDTH + (width / 2), HALF_SCENE_WIDTH - (width / 2)
    
    @staticmethod
    def min_and_max_y(height: float):
        """
        Calculates the minimum and maximum possible 'y' to make the
        object with the provided 'height' fit the manim screen, that
        means to be inside of it and appear in the video.

        The provided 'height' must be a manim height, not a height
        in pixels.

        This method returns the 'min, max' pair of minimum and
        maximum possible y values.
        """
        return -HALF_SCENE_HEIGHT + (height / 2), HALF_SCENE_HEIGHT - (height / 2)

    @classmethod
    def random_position(width: float, height: float):
        """
        Calculate a random position inside the manim screen limits according
        to the provided 'width' and 'height' that must be 

        Provided 'width' and 'height' must be in manim width and height.

        Calculate a random position inside the manim screen limits according
        to the provided 'width' and 'height' that must be of the element to
        position.
        """
        x_min, x_max = ManimPosition.min_and_max_x(width)
        random_x = x_min + (random() * (x_max - x_min))

        y_min, y_max = ManimPosition.min_and_max_y(height)
        random_y = y_min + (random() * (y_max - y_min))
        
        # TODO: Maybe in this class is as easy as return just 'x, y'
        return random_x, random_y

# TODO: Remove this below when refactored and unneeded
def get_random_position(width: float, height: float):
    """
    Returns a random position inside the screen according to the provided element width and
    height to fit in. If you are trying to position a text inside screen limits, you must
    provide text width and height to let this method calculate that random position.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_left_position(width: float, height: float):
    """
    Returns a random position in the upper left corner according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = -(width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_right_position(width: float, height: float):
    """
    Returns a random position in the upper right corner according to the 
    provided element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_center_position(width: float, height: float):
    """
    Returns a random position in the upper center according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH / 2 + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH / 2 - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_upper_position(width: float, height: float):
    """
    Returns a random position in the upper section according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_center_position(width: float, height: float):
    """
    Returns a random position in the center according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH / 2 + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH / 2 - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT / 2 + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT / 2 - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_center_left_position(width: float, height: float):
    """
    Returns a random position in the center left according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = -(width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT / 2 + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT / 2 - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_center_right_position(width: float, height: float):
    """
    Returns a random position in the center right according to the 
    provided element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT / 2 + (height / 2)
    Y_MAXIMUM = HALF_SCENE_HEIGHT / 2 - (height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_left_position(width: float, height: float):
    """
    Returns a random position in the lower left corner according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = -(width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_right_position(width: float, height: float):
    """
    Returns a random position in the lower right corner according to the 
    provided element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_center_position(width: float, height: float):
    """
    Returns a random position in the lower center according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH / 2 + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH / 2 - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

def get_random_lower_position(width: float, height: float):
    """
    Returns a random position in the upper section according to the provided
    element 'width' and 'height' to fit in.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X_MINIMUM = -HALF_SCENE_WIDTH + (width / 2)
    X_MAXIMUM = HALF_SCENE_WIDTH - (width / 2)
    random_x = X_MINIMUM + (random() * (X_MAXIMUM - X_MINIMUM))
    Y_MINIMUM = -HALF_SCENE_HEIGHT + (height / 2)
    Y_MAXIMUM = -(height / 2)
    random_y = Y_MINIMUM + (random() * (Y_MAXIMUM - Y_MINIMUM))

    return {
        'x': random_x,
        'y': random_y
    }

# Exact places
HEIGHT_DISTANCE_FROM_EDGES = height_to_manim_height(10)
WIDTH_DISTANCE_FROM_EDGES = width_to_manim_width(10)
def get_upper_left_position(width: float, height: float):
    """
    Returns the exact position of the upper left corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = -HALF_SCENE_WIDTH + WIDTH_DISTANCE_FROM_EDGES + (width / 2)
    Y = HALF_SCENE_HEIGHT - HEIGHT_DISTANCE_FROM_EDGES - (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_upper_right_position(width: float, height: float):
    """
    Returns the exact position of the upper right corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = HALF_SCENE_WIDTH - WIDTH_DISTANCE_FROM_EDGES - (width / 2)
    Y = HALF_SCENE_HEIGHT - HEIGHT_DISTANCE_FROM_EDGES - (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_upper_center_position(height: float):
    """
    Returns the exact position of the upper center position according 
    to the provided element 'width' and 'height' to fit in and be 
    placed just there.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = 0
    Y = HALF_SCENE_HEIGHT - HEIGHT_DISTANCE_FROM_EDGES - (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_left_position(width: float):
    """
    Returns the exact position of the left side according to
    the provided element 'width' and 'height' to fit in and be placed
    just in that place

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = -HALF_SCENE_WIDTH + WIDTH_DISTANCE_FROM_EDGES + (width / 2)
    Y = 0

    return {
        'x': X,
        'y': Y
    }

def get_center_position():
    """
    Returns the exact position of the center according to the 
    provided element 'width' and 'height' to fit in and be placed
    just in that place

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = 0
    Y = 0

    return {
        'x': X,
        'y': Y
    }

def get_right_position(width: float):
    """
    Returns the exact position of the right side according to
    the provided element 'width' and 'height' to fit in and be placed
    just in that place

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = HALF_SCENE_WIDTH - WIDTH_DISTANCE_FROM_EDGES - (width / 2)
    Y = 0

    return {
        'x': X,
        'y': Y
    }

def get_lower_left_position(width: float, height: float):
    """
    Returns the exact position of the lower left corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = -HALF_SCENE_WIDTH + WIDTH_DISTANCE_FROM_EDGES + (width / 2)
    Y = -HALF_SCENE_HEIGHT + HEIGHT_DISTANCE_FROM_EDGES + (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_lower_right_position(width: float, height: float):
    """
    Returns the exact position of the lower right corner according to
    the provided element 'width' and 'height' to fit in and be placed
    just in the corner.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = HALF_SCENE_WIDTH - WIDTH_DISTANCE_FROM_EDGES - (width / 2)
    Y = -HALF_SCENE_HEIGHT + HEIGHT_DISTANCE_FROM_EDGES + (height / 2)

    return {
        'x': X,
        'y': Y
    }

def get_lower_center_position(height: float):
    """
    Returns the exact position of the lower center position according 
    to the provided element 'width' and 'height' to fit in and be 
    placed just there.

    Provided 'width' and 'height' must be in manim width and height.

    This method returns an object containing 'x' and 'y' random fitting
    coordinates.
    """
    X = 0
    Y = -HALF_SCENE_HEIGHT + HEIGHT_DISTANCE_FROM_EDGES + (height / 2)

    return {
        'x': X,
        'y': Y
    }


"""
Extracted from official documentation:
- https://zulko.github.io/moviepy/getting_started/videoclips.html?highlight=mask#mask-clips

The fundamental difference between masks and standard clips is that standard clips output frames with 3 components (R-G-B) per pixel, comprised between 0 and 255, while a mask has just one composant per pixel, between 0 and 1 (1 indicating a fully visible pixel and 0 a transparent pixel). Seen otherwise, a mask is always in greyscale.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames import VideoFrameExtractor
from yta_general_utils.file.checker import file_is_video_file
from yta_general_utils.file import list
from yta_general_utils.temp import create_custom_temp_filename 
from moviepy.editor import VideoFileClip, ImageClip, concatenate_videoclips, CompositeVideoClip, VideoClip, ImageSequenceClip
from typing import Union

import numpy as np
import cv2


class VideoMaskHandler:
    """
    Class created to simplify and encapsulate the working process
    with moviepy video masks.

    TODO: Maybe move this to VideoParser or to FrameHandler
    """
    @classmethod
    def frame_is_normal_clip_frame(cls, frame: np.ndarray):
        """
        Checks if the provided 'frame' numpy array is recognized as
        a frame of a normal moviepy clip with values between 0 and
        255.

        This numpy array should represent a frame of a clip.
        
        A non-modified clip is '.ndim = 3' and '.dtype = np.uint8'.
        """
        if not isinstance(frame, np.ndarray):
            raise Exception('The provided "clip" parameter is not a numpy array.')
        
        # TODO: We check if it is a normal clip frame (we didn't
        # do '.to_mask()' on it)
        if (
            frame.ndim != 3 or
            frame.dtype != np.uint8 or
            np.all((frame >= 0) & (frame <= 255))
        ):
            # You can convert a clip to mask with '.to_mask()' and it would
            # still be a clip (but with mask numpy types)
            return False
        
    @classmethod
    def frame_is_mask_clip_frame(cls, frame: np.ndarray):
        """
        Checks if the provided 'mask_clip' numpy array is recognized
        as an original moviepy mask clip with values between 0 and 1.
        This numpy array should represent a frame of a mask clip.
        
        A non-modified mask clip is '.ndim = 2' and '.dtype = np.float64'.
        """
        if not isinstance(frame, np.ndarray):
            raise Exception('The provided "mask_clip" parameter is not a numpy array.')
        
        # TODO: We check if it is an original mask clip frame (we didn't
        # do '.to_RGB()' on it)
        if (
            frame.ndim != 2 or
            frame.dtype not in (np.float64, np.float32) or
            np.all((frame >= 0) & (frame <= 1))
        ):
            # You can convert a mask clip to clip with '.to_RGB()' and it would
            # still be a mask clip (but with clip numpy types)
            return False

    @classmethod
    def clip_is_normal_clip(cls, clip):
        """
        Normal clips have frames as numpy arrays with values between
        0 and 255 (where 255 means full of color).

        A non-modified clip is '.ndim = 3' and '.dtype = uint8'.
        """
        # TODO: What if I receive the numpy array (?)

        # TODO: I'm not sure, maybe a CompositeVideoClip is a mask (?)
        if not isinstance(clip, VideoClip):
            # TODO: False or Exception (?)
            return False
        
        return cls.frame_is_normal_clip_frame(clip.get_frame(t = 0))

    @classmethod
    def clip_is_mask_clip(cls, clip):
        """
        Mask clips have frames as numpy arrays with values only between
        0 and 1 (where 1 means completely transparent). 
        
        A non-modified mask is '.ndim = 2' and '.dtype = float64'.
        """
        # TODO: What if I receive the numpy array (?)

        # TODO: I'm not sure, maybe a CompositeVideoClip is a mask (?)
        if not isinstance(clip, VideoClip):
            # TODO: False or Exception (?)
            return False

        return cls.frame_is_mask_clip_frame(clip.get_frame(t = 0))
    
    @classmethod
    def invert_mask_frame(cls, frame: np.ndarray):
        """
        Inverts the values in the provided mask clip 'frame'. A mask
        clip frame is a numpy array with values between 0 and 1, so
        they will be inverted.
        """
        if not cls.frame_is_mask_clip_frame(frame):
            raise Exception('The provided "frame" is not actually a mask frame.')
        
        frame[np.where(frame == 0)] = 1
        frame[np.where(frame == 1)] = 0

        return frame

    @classmethod
    def invert_mask_clip(cls, mask_clip):
        """
        Inverts all the numpy mask pixels of the provided 'mask_clip'
        if a valid mask clip is provided. Only the array of numpy
        mask pixels inverted is returned.
        """
        if not cls.clip_is_mask_clip(mask_clip):
            raise Exception('The provided "mask_clip" parameter is not actually a mask clip.')
        
        inverted_mask_frames = []
        # Mask clip is normalized (one pixel between 0 and 1)
        for frame in mask_clip.iter_frames():
            inverted_mask_frames.append(cls.invert_mask_frame(frame))

        return ImageSequenceClip(inverted_mask_frames, fps = mask_clip.fps, ismask = True)

    @classmethod
    def invert_clip_mask(cls, clip):
        """
        Inverts the provided 'clip' mask. The provided 'clip' must be
        a valid normal moviepy clip with a mask. If no mask, the mask
        will be forced.

        This method returns the provided 'clip' with its mask inverted.

        Careful: if the provided 'clip' doesn't have any mask, it will
        be forced as completely opaque, and when inverted it will be
        completely transparent so the normal 'clip' will never be
        shown.
        """
        clip = VideoParser.to_moviepy(clip, has_mask = True)

        clip.mask = cls.invert_mask_clip(clip.mask)

        return clip

# TODO: Check and refactor this below
def invert_video_mask(video: Union[str, VideoFileClip]):
    """
    This method will invert the provided video mask. Make 
    sure the video you provide has a mask. This method 
    will return a mask containing the original video
    with the mask inverted.

    This method will iterate over each frame and will
    invert the numpy array to invert the opacity by 
    chaging the zeros by ones and the ones by zeros (but
    in 0 - 255 range).

    This is useful to overlay videos that have alpha 
    channels so they can let see the video behind through
    that alpha channel. For example, if your video is 
    just a manim animation with only text in the middle,
    the other pixels in the video will be alpha, so if you
    invert them you will obtain the text-transparent 
    effect that is a pretty artistic one.

    This mask has to be used in a specific way, that is in
    a CompositeVideoClip, in second position, with the
    'use_bgclip = True' flag to be over the other video.
    """
    video = VideoParser.to_moviepy(video, has_mask = True)

    # TODO: This is actually being used to create an effect, so is
    # not only inserting a video mask. It needs to be refactored.

    # TODO: Confirm that this exist or create it if not
    FRAMES_FOLDER_NAME = 'frames_main'
    # This is to avoid memory limit exceeded
    in_memory = False

    # We will invert all frames
    # TODO: Please, try to do this with ffmpeg concatenate
    # images with alpha channel to video because it will be
    # faster but I couldn't in the past
    # TODO: Use FfmpegHandler
    clips = []
    for i in range(int(video.fps * video.duration)):
        mask_frame = VideoFrameExtractor.get_frame_by_frame_number(video.mask, i)
        frame = VideoFrameExtractor.get_frame_by_frame_number(video, i)

        # Invert the mask
        where_0 = np.where(mask_frame == 0)
        where_1 = np.where(mask_frame == 1)
        mask_frame[where_0] = 255
        mask_frame[where_1] = 0

        # Combine the fourth (alpha) channel
        mask_frame = mask_frame[:, :, np.newaxis]
        frame_rgba = np.concatenate((frame, mask_frame), axis = 2)
        if in_memory:
            # Maybe it is not possible to build with ffmpeg (https://superuser.com/a/1706440)
            clips.append(ImageClip(frame_rgba, duration = 1 / 60))
        else:
            tmp_frame_name = create_custom_temp_filename(FRAMES_FOLDER_NAME + '/frame' + str(i).zfill(5) + '.png')
            cv2.imwrite(tmp_frame_name, frame_rgba)
    
    if not in_memory:
        frames_folder = create_custom_temp_filename(FRAMES_FOLDER_NAME + '/')
        images = list(frames_folder, pattern = '*.png')
        # Maybe it is not possible to build with ffmpeg (https://superuser.com/a/1706440)
        for image in images:
            clips.append(ImageClip(image, duration = 1 / 60))

    mask = concatenate_videoclips(clips)

    return mask

# TODO: I think this will be deleted in the future, when refactored
# and when we confirm that here is another way of handling this
def apply_inverted_mask(video: Union[str, VideoFileClip], mask_video: Union[str, VideoFileClip], output_filename: Union[str, None] = None):
    """
    Applies the provided 'mask_video' with its mask inverted
    over the also provided 'video'. This is useful to make
    artistic effects. This methods applies the 
    'invert_video_mask' method to the provided 'mask_video'.
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if not mask_video:
        raise Exception('No "mask_video" provided.')
    
    if isinstance(video, str):
        if not file_is_video_file:
            raise Exception('Provided "video" is not a valid video file.')
        
        video = VideoFileClip(video, has_mask = True)

    if isinstance(mask_video, str):
        if not file_is_video_file:
            raise Exception('Provided "mask_video" is not a valid video file.')
        
        mask_video = VideoFileClip(mask_video, has_mask = True)

    if not mask_video.mask:
        mask_video = VideoFileClip(mask_video.filename, has_mask = True)

    mask_video = invert_video_mask(mask_video)
    # TODO: Handle durations
    final_clip = CompositeVideoClip([video, mask_video.subclip(0, video.duration)], use_bgclip = True)
    final_clip = final_clip.set_audio(video.audio)

    if output_filename:
        final_clip.write_videofile(output_filename)

    return final_clip

# TODO: Look for the way to store the video with the inverted mask locally
# so I can use that video as usual. Normally I download video with alpha
# layers from the Internet or I create them with manim, but now I have to
# handle it manually, so this is a challenge


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, VideoClip, ColorClip
from typing import Union


class ChangeSpeedVideoEffect(VideoEffect):
    """
    This effect changes the speed of the video to fit the requested
    'final_duration', that will accelerate or decelerate the video
    speed.
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], final_duration = None):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        if final_duration is None:
            final_duration = video.duration

        effect_name = VideoEffect.get_moviepy_vfx_effect('speedx')
        parameters = {
            'final_duration': final_duration
        }

        return video.fx(effect_name, **parameters)



from yta_multimedia.video.parser import VideoParser
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.checker.type import variable_is_number_between
from yta_general_utils.file.handler import rename_file
from moviepy.editor import VideoFileClip
from math import floor, ceil
from typing import Union


# TODO: Is this being used (?) It is very useless and non-sense
def rescale_video(video: Union[VideoFileClip, str], output_width: int = 1920, output_height: int = 1080, output_filename: Union[str, None] = None):
    """
    This method was created to rescale videos upper to 1920x1080 or 1080x1920. This is,
    when a 4k video appears, we simplify it to 1080p resolution to work with only that
    resolution. This method returns the VideoFileClip rescaled but also writes it if
    'output_filename' provided.

    The 'output_width' and 'output_height' variables must be [1920 and 1080] or [1080 and
    1920]. Any other pair is not valid.

    This method is used in the script-guided video generation. Please, do not touch =).

    TODO: This method is very strict, so it will need a revision to allow other dimensions
    and keep scaling.
    """
    # We only want to accept 16/9 or 9/16 by now, so:
    if not (output_width == 1920 and output_height == 1080) and not (output_width == 1080 and output_height == 1920):
        print('Sorry, not valid input parameters.')
        return None
    
    video = VideoParser.to_moviepy(video)
    
    SCALE_WIDTH = 16
    SCALE_HEIGHT = 9
    if output_width == 1080 and output_height == 1920:
        SCALE_WIDTH = 9
        SCALE_HEIGHT = 16

    width = video.w
    height = video.h

    # We avoid things like 1927 instead of 1920
    new_width = width - width % SCALE_WIDTH
    new_height = height - height % SCALE_HEIGHT

    proportion = new_width / new_height

    if proportion > (SCALE_WIDTH / SCALE_HEIGHT):
        print('This video has more width than expected. Cropping horizontally.')
        while (new_width / new_height) != (SCALE_WIDTH / SCALE_HEIGHT):
            new_width -= SCALE_WIDTH
    elif proportion < (SCALE_WIDTH / SCALE_HEIGHT):
        print('This video has more height than expected. Cropping vertically.')
        while (new_width / new_height) != (SCALE_WIDTH / SCALE_HEIGHT):
            new_height -= SCALE_HEIGHT

    print('Final: W' + str(new_width) + ' H' + str(new_height))
    videoclip_rescaled = video.crop(x_center = floor(width / 2), y_center = floor(height / 2), width = new_width, height = new_height)
    
    # Force output dimensions
    if new_width != output_width:
        print('Forcing W' + str(output_width) + ' H' + str(output_height))
        videoclip_rescaled = videoclip_rescaled.resize(width = output_width, height = output_height)

    if output_filename:
        # TODO: Check extension
        tmp_video_filename = create_temp_filename('scaled.mp4')
        tmp_audio_filename = create_temp_filename('temp-audio.m4a')
        videoclip_rescaled.write_videofile(tmp_video_filename, codec = 'libx264', audio_codec = 'aac', temp_audiofile = tmp_audio_filename, remove_temp = True)
        rename_file(tmp_video_filename, output_filename, True)

    return videoclip_rescaled

def resize_video(video: Union[VideoFileClip, str], width: int = None, height: int = None, output_filename: Union[str, None] = None):
    """
    Resizes the provided 'video' to the also provided 'width' and
    'height' without scaling. You can provide only one dimension and it
    will calculate the other one keeping the scale.

    This method will return the new video as a VideoFileClip and will write
    a new file only if 'output_filename' is provided.
    """
    # By now I'm limiting it to 1920
    if not variable_is_number_between(width, 1, 1920) or not variable_is_number_between(height, 1, 1920):
        raise Exception('The provided "width" and/or "height" parameters are not valid numbers. They must be between [1, 1920].')
    
    video = VideoParser.to_moviepy(video)

    if width and not height:
        height = ceil((video.h * width) / video.w)
    elif height and not width:
        width = ceil((video.w * height) / video.h)

    # moviepy resize method does not allow odd numbers
    width -= width % 2
    height -= height % 2
    
    video = video.resize(width = width, height = height)

    if output_filename:
        video.write_videofile(output_filename)

    return video


# Useful: https://www.youtube.com/watch?v=kpXU8DM068U
# Maybe buy this? https://sevenstyles.com/p/585-midjourney-prompts-for-christianity-vol3-3779965/
from yta_general_utils.programming.enum import YTAEnum as Enum


class ReligiousPrompt(Enum):
    FRESCO_PAINTING = 'historical fresco, profile view, a timeless depiction of the Last Supper, communion, disciples sharing bread, rustic room, wooden table, warm lamplight, unity, aged effects, 1st century, Pompeii style, earthy oranges and browns, Classical, Hellenistic Classical, ancient, cracked plaster, worn wood, chipped cup, fellowship, fresco technique, biblical meal, authenticity'
    MARY_THE_VIRGIN_SCULPTURE = 'ethereal carving, a gentle touch between Santa Maria and Jesucrist the child, serene, smooth, divine, contemplative, motherly connection, tranquil, altar background, soft diffused light, reverence, subtle marble texture, Renaissance, elegant baroque, delicate, smooth alabaster, timeless, in the style of Michelangelo, muted pastels, profile view, neoclassicism, classical contemporary, refined, polished marble, faith, reverence, sculpture, classical, pure elegance'
    OLD_PAINTING = 'folk art simplicity, side on view, a peaceful nativity with animals and shepherds, joy, innocence and purity, familial bonding, pastoral landscape, thatched roof stable, warm evening glow, renaissance, angelic visitation, subtle moonlight effects, 16th century, Marc Chagall style, pastel hues, impressionism, roman mural, unrefined, coarse straw, smooth porcelain, magi, infant christ, animals'
    ANGELS = 'mystical renaissance art, angelic hierarchies in celestial dance, wings spread, halos glowing, reverence, divinity, cosmic order, high renaissance, sandro botticelli style, aerial view, humanism, mystical renaissance, celestial, ethereal, feathery, reverence, divinity, order, cosmos'
    TOMB = 'mystical fresco painting, wide angle view, empty tomb with light, hope, resurrection and miracle, peaceful garden, stone entrance, morning sunrise, new life, smooth surface, dawn of Christianity, Fra Angelico style, soft earth tones, Byzantine, Futurist, ethereal, stone texture and luminous rays, eternal hope, garden tomb, risen Christ, spiritual awakening'
    WOOD_CROSS = 'minimalist cross, a simple wooden cross against a clear sky, faith and redemption, simplicity, cross on a hill, uncluttered, open sky, bright daylight, timeless, George Tooker style, basic color scheme, distant view, Christian symbolism, wooden cross, unadorned, smooth wood, faith, redemption, cross and sky'
    MANUSCRIPTS = 'impressionistic brushwork, detailed study of angelic figures in illuminated manuscripts, delicate wings, serene faces, golden halos, purity, gentle interaction with human figures, ethereal glow, historical library, soft natural light, sanctity, feathering technique, medieval period, jan van eyck style, pastel blues, medium close up, romanticism, luminous gothic, velvety smooth, textured parchment, subtle ink, grace, devotion, elegance, history, illumination'
    ANGELS_AND_DAEMONS = 'expressive portrayal of angels and demons in christian art, serene depiction of the virgin mary, graceful brushstrokes, symbolic elements, surreal imagination, ethereal atmosphere, baroque style, mythical motif, 17th century, dynamic interplay, heavenly glow, celestial backdrop, soft golden lighting, reverence, caravaggio style, celestial blue, low angle view, symbolism, abstract baroque, velvety texture, feathery wings, divine presence, seraphim, cherubim, motherly love, annunciation'
    QUIET_MONASTERY = 'minimalist monastic, a quiet monastery courtyard with monks in contemplation, soft candlelight casting long shadows, solitude, monks meditating, serene, stone walls, subtle candlelight, medieval, Carthusian style, muted earth tones, low angle view, monasticism, contemplative quietude, austere, textured stone, contemplation, silence, simple living'
    READING_MINIMALIST = 'minimalist vector style, a saint reading scripture, simple shapes, an open book, contemplation, saint and scripture in dialogue, thoughtful, a sparse study room, even lighting, minimalism, vector art, 17th century, saul bass style, earth tones, high angle view, modernism, vector minimalism, geometric, flat texture, sacred reading, studious saint, divine wisdom, scriptural insight'
    METAL_SAINT_RELIC = "intricate metalwork, profile view, on a saint's relic, a chirho mesmerizes, legacy, Christ's monogram, metallic intricacy, saint's tomb, reflective sheen, heritage, etched elegance, Byzantine Empire, Benvenuto Cellini style, oxidized metal, renaissance, Christian metallurgy, etched, reflective, chirho symbol, Christ's legacy, metalwork, historical artistry"
    SACRED_ORTHODOX_FRESCO = 'sacred Orthodox fresco, tilt view, Saint Catherine with the wheel, wisdom, St. Catherine standing, broken wheel, scholarly books, palm of martyrdom, intellect, fresco crackling effects, Orthodox period, Hilandar Monastery style, noble blues and reds, Byzantine Art, Serbian Byzantine, wise, jeweled crown, shattered spokes, open scroll, martyrdom, fresco layering, Saint Catherine, erudition'
    CANDLE_MINIMALIST = "spiritual minimalism, a minimalist interpretation of early christian teachings, soulful connection, divine relationship, sacred ambiance, monastic walls, candlelit shadows, byzantine period, in the style of robert ryman, spiritual colors, high angle view, neoplasticism, robert ryman minimalism, pure, soft, gentle, graceful, minimalism, monochrome painting, robert ryman's spirituality, devotion, monastic minimalism"
    BAROQUE_CRUCIFIX = 'detailed baroque style, an ornate metalwork crucifix in a grand cathedral, intricate details, grandeur, majestic stance, gilded background, rich chandeliers, historical painting, reverence, da vinci style, gold and bronze, high angle, baroque, ornate golden filigree, artistic mastery, renaissance, elaborate metalwork, detailed, metal craftsmanship, art history, traditional richness'
    CROWN_OF_THORNS = 'organic minimalism, a crown of thorns in a desert, sacrifice, isolation, arid desert, barren lands, harsh sunlight, solemnity, empty horizon, rusty browns, ancient times, ana mendieta style, tilt view, land art, organic simplicity, symbolic, coarse sand, twisted metal, sacrifice, desert, organic, minimal, abstract'



# TODO: This below is interesting to randomly choose a method
# but it's been replaced by the YTAEnum random option. I keep
# the code because it is interesting and maybe I move it to
# 'yta_general_utils'
def __is_function_local(object):
    import types

    return isinstance(object, types.FunctionType) and object.__module__ == __name__

def random_prompt():
    """
    Returns a random prompts of the existing ones in this file.
    """
    import inspect
    import sys
    import random

    methods = []
    #print([name for name, value in inspect.getmembers(sys.modules[__name__], predicate = (isinstance(object, types.FunctionType) and object.__module__ == __name__))])
    for name, func in inspect.getmembers(sys.modules[__name__], predicate = __is_function_local):
        if name != 'random_prompt' and name != '__is_function_local':
            methods.append(func)

    # We make a dynamic call to that method and return the random prompt
    return random.choice(methods)()


from moviepy.editor import AudioFileClip, concatenate_audioclips
from yta_multimedia.resources.audio.drive_urls import TYPING_KEYBOARD_3_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL, SILENCE_10_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from yta_multimedia.resources import Resource
from typing import Union


class SoundGenerator:
    # TODO: Move this to a consts.py file
    SILENCE_SOUND_FILENAME = EFFECTS_RESOURCES_FOLDER + 'sounds/silence_10s.mp3'
    TYPING_SOUND_FILENAME = EFFECTS_RESOURCES_FOLDER + 'sounds/typing_keyboard_3s.mp3'

    @classmethod
    def create_silence_audio(cls, duration: float, output_filename: Union[str, None] = None):
        """
        Creates a silence audioclip of the provided 'duration'. If
        'output_filename' is provided, it is also stored locally 
        with that name.
        """
        if not duration:
            raise Exception('No "duration" provided.')
        
        audio_filename = Resource.get(SILENCE_10_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL, cls.SILENCE_SOUND_FILENAME)
        audioclip = AudioFileClip(audio_filename)
        audioclip_duration = audioclip.duration

        # Duration is 10s, so if we need more time we need to concatenate it
        times_to_concat = int(duration / audioclip_duration)
        for i in range(times_to_concat):
            audioclip = concatenate_audioclips([audioclip, audioclip])

        audioclip = audioclip.subclip(times_to_concat * audioclip_duration, times_to_concat * audioclip_duration + duration)

        if output_filename:
            audioclip.write_audiofile(output_filename)

        return audioclip

    @classmethod
    def create_typing_audio(cls, output_filename: Union[str, None] = None):
        """
        Creates a typing audioclip of 3.5 seconds that, if 
        'output_filename' is provided, is stored locally
        with that name.
        """
        audio_filename = Resource.get(TYPING_KEYBOARD_3_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL, cls.TYPING_SOUND_FILENAME)
        audioclip = AudioFileClip(audio_filename)
        silence_audioclip = cls.create_silence_audio(0.5)

        audioclip = concatenate_audioclips([audioclip, silence_audioclip])

        if output_filename:
            audioclip.write_audiofile(output_filename)

        return audioclip



"""
    TODO:

    Hi! Any file and method that is inside the 'experimental'
    folder is not suitable to be actually used with the
    library. The developer is trying the code and making 
    progress but, as the name says, the code is experimental.

    Any use of the code contained in this folder is not the
    developers foul. Thank you for using the library.
"""


from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_center import CoordinateCenter
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner


from yta_multimedia.video.edition.effect.moviepy.position.move.move_arc_position_moviepy_effect import MoveArcPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.move.move_linear_position_moviepy_effect import MoveLinearPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.move.move_sinusoid_position_moviepy_effect import MoveSinusoidPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.move.move_wave_position_moviepy_effect import MoveWavePositionMoviepyEffect


from yta_general_utils.file.handler import rename_file
from yta_general_utils.file.checker import file_is_audio_file
from yta_general_utils.temp import create_temp_filename
from moviepy.editor import AudioFileClip


def crop_audio_file(audio_filename: str, duration: float, output_filename: str):
    """
    Crops the 'audio_filename' provided to the requested 'duration'.

    This method returns the new audio 'output_filename' if valid, or
    False if it was not possible to crop it.
    """
    if not audio_filename:
        return None
    
    if not output_filename:
        return None

    from moviepy.editor import AudioFileClip

    audio_clip = AudioFileClip(audio_filename)

    if audio_clip.duration < duration:
        # TODO: Exception, None, review this
        print('audio is shorter than requested duration')
        return False
    
    audio_clip.set_duration(duration).write_audiofile(output_filename)

    return output_filename

def speedup_audio_file(audio_filename: str, new_duration: int, output_filename: str):
    """
    Receives an audio 'audio_filename' and makes it have the provided
    'new_duration' if shorter than the real one. It will speed up the
    audio to fit that new duration and will store the new audio file
    as 'output_filename'.
    """
    if not audio_filename:
        return None
    
    if not file_is_audio_file(audio_filename):
        return None
    
    audio = AudioFileClip(audio_filename)
    if audio.duration <= new_duration:
        return None
    
    if not output_filename:
        return None

    from pydub import AudioSegment
    from pydub.effects import speedup

    # We calculate audio the speed_up factor 
    speed_factor = audio.duration / new_duration

    # We use a tmp file because input filename could be same as output 
    # TODO: Careful with extension
    tmp_audio_filename = create_temp_filename('tmp_audio_shortened.wav')
    # TODO: What about format?
    sound = AudioSegment.from_file(audio_filename, format = "mp3")
    final = speedup(sound, playback_speed = speed_factor)
    final.export(tmp_audio_filename, format = "mp3")
    # TODO: This is giving 'PermissionError [WindError 5]' when input 
    # and output are the same
    rename_file(tmp_audio_filename, output_filename, True)



from yta_general_utils.file.filename import get_file_filename_without_extension
from subprocess import run


def separate_audio_file_into_vocals_and_instruments(audio_filename: str, output_dirname: str):
    """
    This method will get 'audio_filename' audio file and will split
    it into 2 different files stored in 'output_dirname' directory.
    These 2 files will be 'vocals.wav' and 'accompaniment.wav' that
    are the vocals and the instrumental part, respectively, and 
    will be generated in a folder inside the 'output_dirname' 
    provided that is the 'audio_filename' file name without the
    extension, and will be returned by the function.

    If 'example.wav' provided as 'audio_filename' and 'test' as 
    'output_dirname', this method will create 
    'test/example/vocals.wav' and 'test/example/accompaniment.wav'
    audio files.
    """
    # This is using Spleeter, check this (https://github.com/deezer/spleeter/wiki/2.-Getting-started)
    #
    # This method needs a pre-trained model (that is downloaded from the repo)
    # for each type (2stems, 4stems or 5stems). These are downloaded by default
    # in 'pretrained_models' folder (project root) of the project in which code
    # is being executed
    #
    # You can also do 'spleeter:4stems' and 'spleeter:5stems'
    # 2stems ->   vocals.wav, accompaniment.wav
    # 4stems ->   vocals.wav, drums.wav, bass.wav, piano.wav
    # 5stems ->   vocals.wav, drums.wav, bass.wav, piano.wav and other.wav
    #
    # By now this method is only working for the '2stems' format so it will
    # generate 'vocals.wav' and 'accompaniment.wav' in the provided 
    # 'output_dirname'. We could personalize this in the future.
    if not audio_filename:
        return None
    
    if not output_dirname:
        return None
    
    command_str = 'spleeter separate -p spleeter:2stems -o ' + output_dirname + ' ' + audio_filename

    try:
        run(command_str)
    except:
        return False

    if not output_dirname.endswith('/'):
        output_dirname += '/'

    generated_output_folder = output_dirname + get_file_filename_without_extension(audio_filename)
    
    return generated_output_folder

    # TODO: Please, make it work with python code, this below is not working
    """
    from spleeter.audio.adapter import AudioAdapter
    from spleeter.separator import Separator

    separator = Separator('spleeter:2stems')
    # With this below we read file as numpy array (waveform parameter)
    #from scipy.io.wavfile import read
    #audiowave = read(audio_filename)
    audio_adapter = AudioAdapter.default()
    separator.separate_to_file(audio_descriptor = audio_filename, destination = output_dirname, audio_adapter = audio_adapter)
    separator.join()
    """


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.audio.parser import AudioParser
from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.file.checker import file_is_audio_file, file_is_video_file
from yta_general_utils.file.handler import rename_file
from yta_general_utils.temp import create_temp_filename
from moviepy.editor import VideoClip, AudioClip
from typing import Union

import ffmpeg


def extract_audio_from_video(video: Union[VideoClip, str], output_filename: Union[str, None] = None):
    """
    Returns the audio in the video file provided as 'video_input'. If
    'output_filename' provided, it will write the audio in a file
    with that name.
    """
    video = VideoParser.to_moviepy(video)

    if output_filename:
        # TODO: Check extension, please
        video.audio.write_audiofile(output_filename)

    return video.audio

def set_audio_in_video(video: Union[VideoClip, str], audio: Union[AudioClip, str], output_filename: Union[str, None] = None):
    """
    This method returns a VideoFileClip that is the provided 'video_input' 
    with the also provided 'audio_input' as the unique audio (if valid
    parameters are provided). If 'output_filename' provided, it will
    write the video file with the new audio with that provided name.

    (!) If the input video file and the output file name are the same, you 
    will lose the original as it will be replaced.
    """
    video = VideoParser.to_moviepy(video)
    audio = AudioParser.to_audiofileclip(audio)

    video.audio = audio

    if output_filename:
        # TODO: Check extension, please
        # tmp_output_filename = create_temp_filename('tmp_output.mp4')

        # if variable_is_type(video_input, str) and variable_is_type(audio_input, str) and video_input and audio_input:
        #     # Both are valid str filenames, we will concat with ffmpeg
        #     # TODO: Check if I can get the original filename from an AudioFileClip|VideoFileClip
        #     ffmpeg.concat(ffmpeg.input(video_input), ffmpeg.input(audio_input), v = 1, a = 1).output(tmp_output_filename).run()
        # else:
        #     video_input.write_videofile(tmp_output_filename)
        #
        # rename_file(tmp_output_filename, output_filename, True)

        video.write_videofile(output_filename)

    return video

def set_audio_in_video_ffmpeg(video_filename: str, audio_filename: str, output_filename: str):
    """
    Sets the provided 'audio_filename' in the also provided 'video_filename'
    with the ffmpeg library and creates a new video 'output_filename' that
    is that video with the provided audio.

    TODO: This method need more checkings about extensions, durations, etc.
    """
    if not audio_filename or not file_is_audio_file(audio_filename) or not video_filename or not file_is_video_file(video_filename):
        raise Exception('The provided "audio_filename" and/or "video_filename" are not valid filenames.')
    
    if not output_filename:
        output_filename = create_temp_filename('tmp_audio_ffmpg.mp4')
    
    # TODO: What about longer audio than video (?)
    # TODO: Refactor this below with the FfmpegHandler
    input_video = ffmpeg.input(video_filename)
    input_audio = ffmpeg.input(audio_filename)

    ffmpeg.concat(input_video, input_audio, v = 1, a = 1).output(output_filename).run(overwrite_output = True)


from random import randint


"""
        In place movements (static) effect position functions below
"""
def shake_movement(t, x, y):
    """
    Returns the (x, y) position tuple for the provided 't' of the moviepy
    '.set_position()' method that belongs to a constant shaking process.
    """
    pos = [x, y]
    speed = t * 4
    d = randint(0, 4)

    if 0 == d: #top
        return (pos[0], pos[1] + speed)
    elif 1 == d: #left
        return (pos[0] - speed, pos[1])
    elif 2 == d: #bot
        return (pos[0], pos[1] - speed)
    else: #right
        return (pos[0] + speed, pos[1])
    
def shake_increasing_movement(t, x, y, duration):
    """
    Returns the (x, y) position tuple for the provided 't' of the moviepy
    '.set_position()' method that belongs to a shaking process that
    increases slowly.
    """
    MAX_SHAKE_SPEED = 20
    pos = [x, y]
    # Speed will increase progressively from 0 to MAX_SHAKE_SPEED
    # and this process will last the whole clip duration
    speed = (t / duration) * MAX_SHAKE_SPEED
    d = randint(0, 4)

    if 0 == d: #top
        return (pos[0], pos[1] + speed)
    elif 1 == d: #left
        return (pos[0] - speed, pos[1])
    elif 2 == d: #bot
        return (pos[0], pos[1] - speed)
    else: #right
        return (pos[0] + speed, pos[1])
    
def shake_decreasing_movement(t, x, y, duration):
    """
    Returns the (x, y) position tuple for the provided 't' of the moviepy
    '.set_position()' method that belongs to a shaking process that
    decreases slowly.
    """
    MAX_SHAKE_SPEED = 20
    pos = [x, y]
    # Speed will increase progressively from 0 to MAX_SHAKE_SPEED
    # and this process will last the whole clip duration
    speed = MAX_SHAKE_SPEED - ((t / duration) * MAX_SHAKE_SPEED)
    d = randint(0, 4)

    if 0 == d: #top
        return (pos[0], pos[1] + speed)
    elif 1 == d: #left
        return (pos[0] - speed, pos[1])
    elif 2 == d: #bot
        return (pos[0], pos[1] - speed)
    else: #right
        return (pos[0] + speed, pos[1])
    


from yta_multimedia.greenscreen.custom.image_greenscreen import ImageGreenscreen
from yta_multimedia.greenscreen.custom.video_greenscreen import VideoGreenscreen
from yta_multimedia.resources import Resource
from yta_multimedia.greenscreen.utils import GREENSCREENS_FOLDER
from yta_general_utils.checker.url import url_is_ok
from yta_general_utils.file.checker import is_file, file_is_video_file, file_is_image_file
from yta_general_utils.checker.url import is_google_drive_url
from yta_general_utils.downloader.google_drive import GoogleDriveResource


class Greenscreen:
    """
    This class is an initializer of the ImageGreenscreen and 
    VideoGreenscreen classes so you provide the valid filename
    or Google Drive url of a greenscreen to the 'init' method
    and you get the corresponding ImageGreenscreen or 
    VideoGreenscreen initialized and ready to work.

    If you provide a Google Drive url the system will identify
    the resource and download it only if necessary, as it will
    create locally a unique identifier with the Google Drive ID
    so once it's been downloaded the first time it will be get
    from the local source the next time.

    This will not be possible if a local path is given as the
    system will not be able to identify the resource and the
    greenscreen processing time will be long in any execution.
    """
    @classmethod
    def init(cls, filename_or_google_drive_url: str):
        """
        Parses the provided parameter 'filename_or_google_drive_url' and
        checks if it is a valid filename or a valid Google Drive url. If
        valid, it dowloads the resource (if necessary) and checks if the
        file is an image or a video and returns the corresponding
        ImageGreenscreen or VideoGreenscreen.

        This method will raise an Exception if something goes wrong.
        """
        is_a_file = is_file(filename_or_google_drive_url)
        is_url = False
        filename = filename_or_google_drive_url

        if not is_a_file:
            is_url = url_is_ok(filename_or_google_drive_url)

            if not is_url:
                raise Exception(f'The provided greenscreen "{filename_or_google_drive_url}" parameter is not a file nor a valid url.')
            
            if not is_google_drive_url(filename_or_google_drive_url):
                raise Exception(f'The provided greenscreen "url" parameter "{filename_or_google_drive_url}" is not a valid Google Drive url.')
            
            google_drive_id = GoogleDriveResource(filename_or_google_drive_url).id
            folder = GREENSCREENS_FOLDER + google_drive_id + '/'
            # Extension will be replaced with real one
            filename = Resource.get(filename_or_google_drive_url, folder + 'greenscreen.png')

        # We have a valid filename, lets check if video or image
        if file_is_image_file(filename):
            return ImageGreenscreen(filename)
        elif file_is_video_file(filename):
            return VideoGreenscreen(filename)
        else:
            raise Exception(f'The greenscreen file "{filename}" is not a valid image or video file.')
    


from yta_general_utils.file import checker as file_checker
from yta_general_utils.file import enums as file_enums
from yta_general_utils.file.filename import filename_is_type
from moviepy.editor import VideoFileClip, CompositeVideoClip, ColorClip, ImageClip
from typing import Union


class VideoParser:
    """
    Class to simplify the way we parse video parameters.
    """
    @classmethod
    def to_moviepy(cls, video: Union[str, VideoFileClip, CompositeVideoClip, ColorClip, ImageClip], has_mask: bool = False):
        """
        This method is a helper to turn the provided 'video' to a moviepy
        video type. If it is any of the moviepy video types specified in
        method declaration, it will be returned like that. If not, it will
        be load as a VideoFileClip if possible, or will raise an Exception
        if not.
        """
        if not video:
            raise Exception('No "video" provided.')
        
        if not isinstance(video, str) and not isinstance(video, VideoFileClip) and not isinstance(video, CompositeVideoClip) and not isinstance(video, ColorClip) and not isinstance(video, ImageClip):
            raise Exception('The "video" parameter provided is not a valid type. Check valid types in method declaration.')
        
        if isinstance(video, str):
            if not filename_is_type(video, file_enums.FileType.VIDEO):
                raise Exception('The "video" parameter provided is not a valid video filename.')
            
            if not file_checker.file_is_video_file(video):
                raise Exception('The "video" parameter is not a valid video file.')
            
            video = VideoFileClip(video, has_mask = has_mask)

        # TODO: Maybe '.add_mask()' (?)

        return video
